//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator

package gecko;
import gecko.zoo.*;
import gecko.woods.FragSite;

import java.awt.Color;
import java.util.Vector;
import stats.*;
import gecko.statglue.*;
import courseware.util.*;
import courseware.plot.*;
import courseware.jfcdep.*;
import java.lang.reflect.*;

public class CompositeSpecies {
  private Vector members; // of Species
  private String name;
  private Color color;
  private int colorId;

  private FragStatBuckets fragStats = null;

  public CompositeSpecies(String name, Color c) {
    members = new Vector(1,1);
    this.name = name;
    color = c;
    colorId = QuadDisplay.setColor(name, c);
  }

  public Vector getMembers() { return members;}

  public String getName() { return name;}
  public void setName(String n) { name = n;}

  public void add(Species s){
    if (!members.contains(s))
      members.addElement(s);
  }
  
  public void empty() {members = new Vector(1,1);}

  public void observePopulation() {
    PopBroker[] brokers = new PopBroker[members.size()];
    for (int i=0; i<members.size(); i++)
      brokers[i] = ( (Species)members.elementAt(i) ).getPopBroker();
    observeXTimeCommon(Globals.POP_PLOT_NAME, brokers, name, color, false);
  }
  public void observeBiomass() {
    BiomassBroker[] brokers = new BiomassBroker[members.size()];
    for (int i=0; i<members.size(); i++)
      brokers[i] = ( (Species)members.elementAt(i) ).getBiomassBroker();
    observeXTimeCommon(Globals.BIO_PLOT_NAME, brokers, name, color, false);
  }

  /** Hook broker onto given plot w/ time on x-axis. */
  private void observeXTimeCommon(String plotname, StatSourceBroker[] brokers,
                                  String tracename, Color color, boolean legendFlag)
  {
    MetaPlot mp = Globals.getStats().getMetaPlot(plotname);
    if (mp.hasTrace(tracename)) {
      return;
    }
    SumTrace st = new SumTrace(tracename, color, new Globals());
    for (int i=0; i<brokers.length; i++)
      st.addY( brokers[i] );
    mp.addTrace(st);
    mp.setNeedLegend(legendFlag);
    Globals.getUI().makeStatDumpers(mp);
  }
  /** Hook broker onto given plot w/ time on x-axis. */
  private void observeXTimeSingle(String plotname, StatSourceBroker broker,
                                  String tracename, Color color, boolean legendFlag)
  {
    MetaPlot mp = Globals.getStats().getMetaPlot(plotname);
    if (mp.hasTrace(tracename)) {
      return;
    }
    StatTrace st = new StatTrace(tracename, broker);
    st.setXTime(new Globals());
    st.setColor(color);
    mp.addTrace(st);
    mp.setNeedLegend(legendFlag);
    Globals.getUI().makeStatDumpers(mp);
  }

  private Param makeQuery(String prompt, boolean isboolean,
                           int ival, boolean bval, String methodName)
  {
    Param p;
    Class[] signature = new Class[1];
    Object value;
    if (isboolean) {
      signature[0] = boolean.class;
      value = new Boolean(bval);
    } else {
      signature[0] = int.class;
      value = new Integer(ival);
    }
    Object[] args = new Object[]{value};

    try {
      Method method = this.getClass().getMethod(methodName, signature);
      p = new Param(prompt, args, this, method, true);
    } catch (Exception e) {
      System.err.println("Trouble building fragstat menu");
      return null;
    }
    return p;
  }

  private Param makeGloss(String text)
  {
    Param p = new Param();
    p.setName(text);
    p.setVisible(true);
    return p;
  }

  /** observeFragments is quiet. This one is to put on u.i. and includes
   *  settings quiz. There's a lot of resetting checks needed to make this
   *  work well. Let's skip it for now...we DO know by now if there's a FragSite.
   */
  public void observeFragmentQuiz() {
    // build the quiz
    ParamList quiz = new ParamList();
    quiz.setVisible(true);
    quiz.setName(name + " Fragment Stat Settings");
    quiz.setOwner(this);
    try {
      quiz.setMethod("observeFragments"); // here's the callback
    } catch (Exception e) {
      System.err.println("failed to set callback on observeFrag input dialog");
    }

//    quiz.add(makeGloss("- defining \"fragments\" (grid/sites) -"));
    // if put "use grids yes/no" -after- set x and y, it will work....
    quiz.add(makeQuery("number of grids in x",
      false, 0, false, "setObserveFragmentsGridX" ));
    quiz.add(makeQuery("number of grids in y",
      false, 0, false, "setObserveFragmentsGridY" ));
    quiz.add(makeQuery("use grids (instead of site frags)",
      true, 0, false, "setObserveFragmentsUseGrid" ));
//    quiz.add(makeGloss("- choosing which traces to display -"));
//    quiz.add(makeGloss("- 0 species, 1 species... -"));
    quiz.add(makeQuery("trace by fragments with n species",
      true, 0, true, "setObserveFragmentsByNOF" ));
//    quiz.add(makeGloss("- Hemlock, Maple, ... -"));
    quiz.add(makeQuery("trace by fragments with each species",
      true, 0, false, "setObserveFragmentsBySpecies" ));

    // popup quiz and then do nothing--the quiz itself calls observeFragments
    ParamDialog pd = new ParamDialog(quiz.name, quiz);
    pd.show(ParamDialog.WITH_BOTH_BUTTONS);
  }

  /** FragStats. Complicated by the fact it has optional arguments. Is not in
   *  CompositeSpecies (though only works for CompositeSpecies) because need the
   *  misc. routines here, and Composite doesn't subclass Species.
   */
  public void observeFragments() {
    // test what we can--alas can't check whether there are frags to observe...
    // 1. There are members and they're all QuadCircleSpecies.
    if (members.size() < 2) {
      Globals.log("Skipping observeFragments: composite " + name + " has too few species.");
      return;
    }
    for (int i=0; i<members.size(); i++) {
      Species member = (Species)members.elementAt(i);
      if ( !(member.getProgenitor() instanceof QuadCircleSpecies) ) {
        Globals.log("Skipping observeFragments: composite " + name +
          ", has non-Circle member " + member.getName());
        return;
      }
    }

    // 2. get fragstat machinery ready to roll
    checkFragStats();

    // 3. setup the plots
    FragBroker[] brokers = fragStats.getBrokers();
    for (int i=0; i<brokers.length; i++) {
      observeXTimeSingle(fragStats.getTitle(), brokers[i],
        fragStats.getName(brokers[i].getStat()),
        fragStats.getColor(brokers[i].getStat()), true);
    }

    // 4. Tell the individual members to notifyAlive us.
    for (int i=0; i<members.size(); i++) {
      Species s = (Species)members.elementAt(i);
      s.addFragStats(fragStats);
    }
  }

  /** Prepare fragstats data structures.*/
  private void checkFragStats() {
    if (fragStats == null) {
      fragStats = new FragStatBuckets(this);
    } else {
      fragStats.reinit();
    }
  }

  public void setObserveFragmentsGrid(int x, int y) {
    checkFragStats();
    fragStats.setNumGridX(x);
    fragStats.setNumGridY(y);
  }
  public void setObserveFragmentsGridX(int x) {
    checkFragStats();
    fragStats.setNumGridX(x);
  }
  public void setObserveFragmentsUseGrid(boolean b) {
    if (b) { // do nothing
    } else {
      setObserveFragmentsGridX(0);
      setObserveFragmentsGridY(0);
    }
  }
  public void setObserveFragmentsGridY(int y) {
    checkFragStats();
    fragStats.setNumGridY(y);
  }
  public void setObserveFragmentsMode(int n) {
    checkFragStats();
    fragStats.setMode(n);
  }           
  public void setObserveFragmentsByNOF(boolean b) {
    checkFragStats();
    int modewas = fragStats.getMode();
    if (((modewas == FragStatBuckets.BY_BOTH) || (modewas == FragStatBuckets.BY_NOF) )) {
      if (!b)
        fragStats.setMode(FragStatBuckets.BY_SPECIES);
      return;
    } else { // was by species
      if (b)
        fragStats.setMode(FragStatBuckets.BY_BOTH);
    }
  }
  public void setObserveFragmentsBySpecies(boolean b) {
    checkFragStats();
    int modewas = fragStats.getMode();
    if (((modewas == FragStatBuckets.BY_BOTH) || (modewas == FragStatBuckets.BY_SPECIES) )) {
      if (!b)
        fragStats.setMode(FragStatBuckets.BY_NOF);
      return;
    } else { // was by NOF
      if (b)
        fragStats.setMode(FragStatBuckets.BY_BOTH);
    }
  }

  private ExecParam makeEP(String observeWhat) {
    ExecParam ep;
    try {
      Method observeM = this.getClass().getMethod("observe" + observeWhat,
        Param.NO_ARGS_SIGNATURE);
      ep = new ExecParam(observeWhat, Param.NO_ARGS, this, observeM);
    } catch (Exception e) {
      System.err.println("Couldn't observe " + name + " statistic " + observeWhat);
      return null;
    }
    return ep;
  }

  /** Build list of ExecParam's to turn on stats for one composite. */
  public ParamList getStatParamList() {
    ParamList answer = new ParamList();
    answer.setName(name + " Stats");
    answer.add(makeEP("Population"));
    answer.add(makeEP("Biomass"));
    answer.add(makeEP("Fertility"));
    answer.add(makeEP("Mortality"));
    answer.add(makeEP("Stages"));
    answer.add(makeEP("Fragments"));
    answer.add(makeEP("BiomassTransfers"));
    answer.add(makeEP("BiomassRates"));
    answer.add(makeEP("BiomassGrowth"));
    answer.add(makeEP("BiomassTheft"));
    return answer;
  }


  /** [][] is [members][stats] */
  public BiomassBucketBroker[][] getBioTransBrokers(){
    BiomassBucketBroker[][] brokers = new BiomassBucketBroker[members.size()][];
    for (int i=0; i<members.size(); i++)
      brokers[i] = ( (Species)members.elementAt(i) ).getBioTransBrokers();
    return brokers;
  }

  public BiomassBucketBroker[][] getBioRateBrokers(){
    BiomassBucketBroker[][] brokers = new BiomassBucketBroker[members.size()][];
    for (int i=0; i<members.size(); i++)
      brokers[i] = ( (Species)members.elementAt(i) ).getBioRateBrokers();

    // wrinkle--need a copy of these brokers, with weighting set on, for avg'ing
    for (int i=0; i<members.size(); i++) {
      for (int j=0; j<brokers[i].length; j++) {
        BiomassBucketBroker bbb = new BiomassBucketBroker(brokers[i][j]);
        bbb.setWeightFlag(true);
        brokers[i][j] = bbb;
      }
    }
    return brokers;
  }

  public void observeBiomassTransfers() {
    observeAllTransferCommon(Globals.BIO_TRANSFERS_PLOT_NAME, getBioTransBrokers(), true);
  }
  public void observeBiomassRates() {
    observeAllTransferCommon(Globals.BIO_RATE_PLOT_NAME, getBioRateBrokers(), false);
  }

  /** Transfers require sum-of-buckets, rates require weighted-avg-of-buckets. */
  private void observeAllTransferCommon(String plotsuffix, BiomassBucketBroker[][] brokers,
    boolean sumFlag)
  {
    String[] traceNames  = BiomassTransfers.getNames();
    Color[]  traceColors = BiomassTransfers.getColors();
    Timer timer = new Globals();
    SumTrace st;

    MetaPlot mp = Globals.getStats().getMetaPlot(name + " " + plotsuffix);
    if (mp.hasTrace(traceNames[BiomassTransfers.NET_GROWTH])) {
      return;
    }

    // make and register the sum traces, one for each stat to watch
    for (int i=0; i < BiomassTransfers.getNumStats(); i++) {
      if ( (i != BiomassTransfers.BIOMASS) && (i != BiomassTransfers.LAST_NET_GROWTH)
           && (i != BiomassTransfers.NET_RATIO) ) {
        if (sumFlag)
          st = new SumTrace(traceNames[i], traceColors[i], timer);
        else
          st = new AvgTrace(traceNames[i], traceColors[i], timer);

        // for each member of composite, add broker to each sumtrace
        for (int j=0; j<brokers.length; j++)
          st.addY(brokers[j][i]);

        // add completed sumtrace to the metaplot
        mp.addTrace(st);
      }
    }

    mp.setNeedLegend(true);
    Globals.getUI().makeStatDumpers(mp);
  }
  /** Stat vs. biomass trace. */
  private void observeSnangleCommon(String plotsuffix, int which,
                                    BiomassBucketBroker[][] brokers)
  {
    SumTrace st; // summed values on y and x axes

    MetaPlot mp = Globals.getStats().getMetaPlot(name + " " + plotsuffix);
    if (mp.hasTrace(name)) {
      return;
    }

    // for each member of composite, add broker to sumtrace
    st = new SumTrace(name, color);
    // both x and y values of trace are sums
    for (int j=0; j<brokers.length; j++) {
      st.addY(brokers[j][which]);
      st.addX(brokers[j][BiomassTransfers.BIOMASS]);
    }
    // add completed sumtrace to the metaplot
    mp.addTrace(st);
    Globals.getUI().makeStatDumpers(mp);
  }

  public void observeBiomassTheft() {
    observeSnangleCommon(Globals.BIO_THEFT_PLOT_NAME, BiomassTransfers.LOSS_STEAL,
                         getBioTransBrokers());
  }
  public void observeBiomassGrowth() {
    observeSnangleCommon(Globals.BIO_GROWTH_PLOT_NAME, BiomassTransfers.GROWTH,
                         getBioTransBrokers());
  }

  /** Note the time component of fertility rates is "per statcycle". */
  public void observeFertility() {
    if (!stageObservableCheck()) return;
    observeStageCommon(Globals.STAGE_FERT_PLOT_NAME, getFertBrokers());
  }
  /** Note the time component of fertility rates is "per statcycle".
   *  Mortality is deadpercent of the members who were in the stage at
   * the beginning of the cycle, so it can't exceed 1.0.
   */
  public void observeMortality() {
    if (!stageObservableCheck()) return;
    observeStageCommon(Globals.STAGE_MORT_PLOT_NAME, getMortBrokers());
  }
  /** Stage 0 is alwasy the stage newborns go into. These stats are not
   *  divided by statcycle, so implicitly, it's per statcycle.
   */
  public void observeStages() {
    if (!stageObservableCheck()) return;
    observeStageCommon(Globals.STAGE_PLOT_NAME, getStageBrokers());
  }

  private boolean stageObservableCheck() {
    if (members.size() == 0) {
      Globals.getUI().log("WARNING: can't observeFertility/Mortality/Stages before composite has members.");
      return false;
    }
    Speciesable oneMember = (Speciesable) ((Species)members.elementAt(0)).getProgenitor();
    if (oneMember instanceof StageObservable)
      return true;
    else {
      Globals.getUI().log("WARNING: observeFertility/Mortality/Stages ignored--species doesn't support it.");
      return false;
    }
  }

  private void observeStageCommon(String plotsuffix, StageBucketBroker[][] brokers) {
    String plotname = name + " " + plotsuffix;
    Color[] colors = StageStatBuckets.getColors();
    Timer t = new Globals();
    if ( !( ( (Species)members.elementAt(0) ).stageObservableCheck() ) )
      System.err.println("CompositeSpecies " + name + " isn't stage observable");
    StageObserver oneObserver = ((Species)members.elementAt(0)).getStageObserver();
    MetaPlot mp = Globals.getStats().getMetaPlot(plotname);
    if (mp.hasTrace(oneObserver.getStageName(0))) {
      return;
    }
    AvgTrace st = null;
    for (int i=0; i<brokers[0].length; i++) {
      st = new AvgTrace(oneObserver.getStageName(i), colors[i], t);
      for (int j=0; j<brokers.length; j++)
        st.addY( brokers[j][i] );
      mp.addTrace(st);
    }
    mp.setNeedLegend(true);
    Globals.getUI().makeStatDumpers(mp);
  }

  /** wrinkle--need a -copy- of these brokers, with weighting set on, for avg'ing */
  private StageBucketBroker[][] weightBrokers(StageBucketBroker[][] brokers) {
    for (int i=0; i<brokers.length; i++) {
      for (int j=0; j<brokers[i].length; j++) {
        StageBucketBroker bbb = new StageBucketBroker(brokers[i][j]);
        bbb.setWeighted(true);
        brokers[i][j] = bbb;
      }
    }
    return brokers;
  }

  public StageBucketBroker[][] getStageBrokers(){
    StageBucketBroker[][] brokers = new StageBucketBroker[members.size()][];
    for (int i=0; i<brokers.length; i++)
      brokers[i] = ( (Species)members.elementAt(i) ).getStageBrokers();
    return weightBrokers(brokers);
  }
  public StageBucketBroker[][] getMortBrokers(){
    StageBucketBroker[][] brokers = new StageBucketBroker[members.size()][];
    for (int i=0; i<members.size(); i++)
      brokers[i] = ( (Species)members.elementAt(i) ).getMortBrokers();
    return weightBrokers(brokers);
  }
  public StageBucketBroker[][] getFertBrokers(){
    StageBucketBroker[][] brokers = new StageBucketBroker[members.size()][];
    for (int i=0; i<members.size(); i++)
      brokers[i] = ( (Species)members.elementAt(i) ).getFertBrokers();
    return weightBrokers(brokers);
  }

}
