//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator
/**
 * 990421, vfb, original Java
 * Translated from the Swarm Objective-C Gecko source, version 0.51
 */
/** This Quad produces a single resource every time step, evenly distributed
 *  across its area.
 */

package gecko;
import gecko.zoo.*;

public class FeedQuad extends ResourceQuad {
  private double prodfunc;	// production per unit area per step.
  private double feedfunc;	// current feed per unit area.
  private double reserveLimit; // cap
  private Mcr clipm = new Mcr(); // we clip a lot, keep the clipper around

  public FeedQuad() {
    super();
    setType(QUAD_FEED);
  }

  public void setFrom(FeedQuad other) {
    super.setFrom(other);
    prodfunc = other.prodfunc;
    feedfunc = other.feedfunc;
    reserveLimit = other.reserveLimit;
  }

  public FeedQuad copymom() {
    FeedQuad newbie = new FeedQuad();
    newbie.setFrom(this);
    return newbie;
  }

  /** This is how feedquad's get "born" by Site. They don't die. */
  public Object newCopy() {
    getSite().notifyBirth();
    return this.copymom();
  }

  public void setProductionRate(double pr) {prodfunc = pr;}
  public double getProductionRate() {return prodfunc;}
  public void setLimit(double lim) {reserveLimit = lim;}
  public double getLimit() {return reserveLimit;}
  public void set(int whatResource, double productionRate, double limit) {
    setReservoir(0);
    setStartReservoir(0);
    setResource(whatResource);
    prodfunc = productionRate;
    reserveLimit = limit;
  }

  /** Adds an amount, not a density...and only if point inside or on mcr. */
  public void addResource(double amount, double x, double y) {
    if (isInsideOrOn(x, y))
      reservoir += amount;
  }

  public boolean produces(int res) {return (resource == res);}
  public int getResource() { return resource; }

  public double curFeed() {return feedfunc;}

  /** Edge of this quad qualifies as a competitor for setting direction
   * that eater wanna grow. "Resource" is ignored.  We feed what we got
   * over that area, times percent. So if we have 1.0 per unit area, and
   * percent is .5, and area of intersection eater->feedQuad = .5,
   * we feed the eater .25 (1.0 x .5 x .5).
   */
  public double feedCritter(Critter eater, double percent, Mcr area) {
    double meal = 0;
	  double lostmeal = 0;
    double wx, wy;
    int crossflag; // an MCR_ relation.

    crossflag = this.intersectMcr(area, clipm);
    if (eater.eats(resource)) {
      meal = feedfunc * percent * clipm.getArea();
      reservoir -= meal;
      eater.feed(resource, meal);
    }

    // set wanna if edge cost this critter resources
    if (crossflag == MCR_INTERSECT) {
      lostmeal = area.getArea() - clipm.getArea();
      wx = clipm.getCenterX();
      wy = clipm.getCenterY();
      wx -= eater.getX();
      wy -= eater.getY();
      eater.addDirection(wx, wy, lostmeal);
    }

    fed += meal;
    return meal;
  }

  public void stepResources() {
    double area = getArea();
    if (reservoir < 0.0)
        reservoir = 0.0;

    reservoir += prodfunc * area;

    if (reservoir > (reserveLimit * area))
        reservoir = (reserveLimit * area);

    feedfunc = reservoir / area;
    startres = reservoir;
    zeroFed();
  }

  public void paint(QuadDisplay qd) {
    qd.drawRect(getSite().getColor(), lx, ly, ux, uy, false);
  }

  public String toString() {
    // dunno how to print address of Object, so use hashCode as good-enough id

    return "FeedQuad[resource=" + resource + ",feedfunc=" + feedfunc

            + ",prodfunc=" + prodfunc + ",limit=" + reserveLimit

            +","+ super.toString() + "]";

  }

}
