//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator
/**
 * 990329, vfb, original Java
 * Translated from the Swarm Objective-C Gecko source, version 0.51
 * 990628, vfb: optional legend
 */
/** Draw Quad tree.
 */
package gecko;

import courseware.util.*;
import courseware.plot.*;
import courseware.jfcdep.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;
import gecko.zoo.*;
import gecko.woods.*;

public class QuadDisplay extends BufferCanvas {
  private Quad quadroot;
  private Mcr truewin;      // that which is a "fit" to window
  private Mcr curwin;       // that zoomed du jour, usually truewin
  private double ppw;				// pixels per world factor
  private boolean displayQuads = false; // visual debug
  private boolean showLegend   = false;
  private Vector toLegend;
  private static Palette palette;  // for data
  private static Palette qpalette; // for quads
  private static int lastQuadColor = 5;
  private static final float circleThickness = DEFAULT_CIRCLE_THICKNESS;
  private static boolean amDrawing = false;

  // save these so can remove/restore them
  private MouseAdapter mousePressedListener;
  private MouseAdapter mouseReleasedListener;
  private MouseMotionAdapter mouseDraggedListener;
  private boolean mouseZoomEnabled = false;
  private Rectangle selection = new Rectangle();
  private boolean paintBetweenXorFlag = false; // xorSelection/paint contention broker

  static {
    qpalette = new Palette(lastQuadColor+1, Palette.BRIGHT, Palette.bluebegin, 1.0f, 0f);
    palette = new Palette();
  }


  public QuadDisplay(Quad root) {
    super();
    quadroot = root;
    truewin = new Mcr();
    curwin = truewin;
    clearLegend();
    initZoom();
  }

  private void initZoom() {
    // add mouse listeners for zooming the display
    // need to save these or no-one can remove them to alter
    // mouse meaning.
    mousePressedListener = new MouseAdapter() {
         public void mousePressed(MouseEvent e) {
             startSelection (e.getX(),e.getY());
         }
    };
    this.addMouseListener(mousePressedListener);

    mouseDraggedListener = new MouseMotionAdapter() {
         public void mouseDragged(MouseEvent e) {
            synchronized (selection) {
             xorSelection(selection.x, selection.y, selection.width, selection.height);
             selection.setSize(e.getX()-selection.x,
                               e.getY()-selection.y);
             xorSelection(selection.x, selection.y, selection.width, selection.height);
            }
         }
    } ;
    this.addMouseMotionListener(mouseDraggedListener);

    mouseReleasedListener = new MouseAdapter() {
        public void mouseReleased(MouseEvent e) {
            endSelection (e.getX(),e.getY());
            zoom();
            repaint();
        }
    } ;
    this.addMouseListener(mouseReleasedListener);
    mouseZoomEnabled = true;
  }

  public void setZoom(boolean onFlag) {
    // do NOT want multiple listeners set or may get double-zooms
    if (mouseZoomEnabled && onFlag)
        return;
    if (!mouseZoomEnabled && !onFlag)
        return;
    if (onFlag) {
        this.addMouseListener(mousePressedListener);
        this.addMouseMotionListener(mouseDraggedListener);
        this.addMouseListener(mouseReleasedListener);
    } else {
        this.removeMouseListener(mousePressedListener);
        this.removeMouseMotionListener(mouseDraggedListener);
        this.removeMouseListener(mouseReleasedListener);
    }
    mouseZoomEnabled = onFlag;
  }

  /** The user has clicked over the plot. Mark Zoom Area */

  /* bug fix--how to get a simple rectangle, instead of it getting
   * drawn inside out or something....
   */
  public void xorSelection(int sx, int sy, int sw, int sh) {
    if (!paintBetweenXorFlag) {
      Graphics g = getGraphics();
      g.setXORMode (Color.cyan);
      int x1, y1, w, h;
      if (sw >= 0) {
          x1 = sx;
          w  = sw;
      } else {
          w  = Math.abs(sw);
          x1 = sx - w;
      }
      if (sh >= 0) {
          y1 = sy;
          h  = sh;
      } else {
          h = Math.abs(sh);
          y1 = sy - h;
      }
      g.drawRect (x1, y1, w, h);
      g.setPaintMode ();
    }
   paintBetweenXorFlag = false;
  }

  private void startSelection (int x, int y)
  {
    paintBetweenXorFlag = false;
    Graphics g = getGraphics();
    selection = new Rectangle (x,y,0,0);
  }

  private void endSelection (int x,int y)
  {
    xorSelection(selection.x, selection.y, selection.width, selection.height);
    selection.setSize(x-selection.x,y-selection.y);
    paintBetweenXorFlag = false;
  }

  static public boolean isDrawing() { return amDrawing;}

  public void setDisplayQuads(boolean b) { displayQuads = b;}

  public void setRoot(Quad q) {quadroot = q;}

  public void clearLegend() {
    showLegend = false;
    toLegend = new Vector(0,5);
  }
  public void addToLegend(String name) {
    int colorid = palette.findName(name);
    if (colorid >= 0) {
      showLegend = true;
      toLegend.addElement(new Integer(colorid));
    }
  }

  /** Returns color index to use hereafter in referring to this color.
   *  If finds this name already, resets the color. If finds color already,
   *  resets the name. Else, adds to the palette.
   */
  public static int setColor(String name, Color c) {
    for (int i=0; i<palette.getNumColors(); i++) {
      if (palette.getColorObj(i).name.equals(name)) {
        palette.getColorObj(i).color = c;
        return i;
      }
      if ( palette.getColorObj(i).color == c ) {
        palette.getColorObj(i).name = name;
        return i;
      }
    }
    int i = palette.addColor(c, name);
    return i;
  }

  /** version that requires no name, does no checking */
  public static int setColor(Color c) {
    return palette.addColor(c, " ");
  }

  public static Color getColor(int i) {
    return palette.getColorObj(i).color;
  }

  private void setPPW() {
    double facx = ((double)getSize().width) / (curwin.ux - curwin.lx);
    double facy = ((double)getSize().height) / (curwin.uy - curwin.ly);
    ppw = facx < facy ? facx : facy;
  }


  synchronized public void setSize(Dimension d)   {
    super.setSize(d);
    setPPW();
  }

  synchronized public void setBounds(int x, int y, int width, int height)   {
    super.setBounds(x,y,width,height);
    setPPW();
  }

  public float w2pfX(double wx) {return( (float) ((wx - curwin.lx) * ppw) );}
  public int w2pX(double wx) {return( (int) ((wx - curwin.lx) * ppw) );}

  public float w2pfY(double wy) {return( (float) ((curwin.uy - wy) * ppw) );}
  public int w2pY(double wy) {return( (int) ((curwin.uy - wy) * ppw) );}
  public float w2pf(double dist) {return( (float) (dist * ppw) );}
  public int w2p(double dist) {return( (int) (dist * ppw) );}

  public double p2wX(int px) {return ((double)px/(double)ppw) + curwin.lx;}
  public double p2wY(int py) {return curwin.uy - ((double)py/(double)ppw);}
  public double p2w(int pdist) {return ((double)pdist/(double)ppw);}

  public void drawCircle(int color, double x, double y, double r, boolean fill) {
    int px0, py0, prf2;
    px0 = w2pX(x - r);
    py0 = w2pY(y + r);
    prf2 = w2p(2*r);
    if (prf2 < 4) {
      prf2 = 4;
    }
    if (fill)
      super.fillCircle(px0, py0, prf2, palette.getColor(color));
    else
      super.drawCircle(px0, py0, prf2, circleThickness, palette.getColor(color));
  }

  /** Since ovals aren't part of the normal rect/circle schemed, don't require
   *  them to be declared to QuadDisplay, so use Color instead of index. */
  public void drawOval(Color color, double lx, double ly, double ux, double uy, boolean fill) {
    Graphics graphics = getOffscreen();
    int px = w2pX( lx );
    int py = w2pY( uy );
    int pw = w2p(ux - lx);
    int ph = w2p(uy - ly);
    if (ph < 3) ph = 3; // saves eyestrain....
    if (pw < 3) pw = 3; // saves eyestrain....
    synchronized (graphics) {
      graphics.setColor(color);
      graphics.drawOval(px, py, pw, ph);
      if (fill) {
        graphics.fillOval(px+1, py+1, pw-2, ph-2);
      } else {
        graphics.drawOval(px+1, py+1, pw-2, ph-2);
      }
    }
  }

  /** x1,y1=lowerleft, x2,y2=upperright */
  public void drawRect(Color color, double x1, double y1, double x2, double y2, boolean fill) {
    Graphics graphics = getOffscreen();
    int px1 = w2pX(x1);
    int py1 = w2pY(y2);
    int width = w2p(x2 - x1);
    int height = w2p(y2 - y1);
    synchronized (graphics) {
      graphics.setColor(color);
      if (fill) {
        graphics.drawRect(px1, py1, width, height);
        graphics.fillRect(px1, py1, width, height);
      } else
        graphics.drawRect(px1, py1, width, height);
    }
  }

  /** Draw string w/ origin at x1, y1 only if it fits within box. */
  public void drawString(Color color, String s, double x1, double y1, double x2, double y2) {
    int px1 = w2pX(x1);
    int py1 = w2pY(y1);
    int width = w2p(x2 - x1);
    int height = w2p(y2 - y1);
    int swidth = getStringWidth(s) - 4; // because that routine adds 4
    int sheight = getStringHeight();
    if ((swidth > width) || (sheight > height))
      return;
    px1 += 0.5 * (width - swidth);
    super.drawString(color, s, px1, py1);
  }



  /** Draw the whole quadtree. */
  public void drawTree() {
    amDrawing = true;
    drawTree(0, quadroot);
    amDrawing = false;
  }

  /** The recursion, with quad color...*/
  private void drawTree(int qcolor, Quad root) {
  // These trees can be vast--don't do both directions recursively....
  // And optimize every bit possible.
    Quad thislevel = root;
    QuadCircleSpecies c;

    if (thislevel == null)
      return;
    if (qcolor > lastQuadColor) qcolor = lastQuadColor;

    do {
      if (thislevel.child != null)
        drawTree(qcolor+1, thislevel.child);
      if (displayQuads) {
        drawRect(qpalette.getColor(qcolor), thislevel.lx, thislevel.ly, thislevel.ux, thislevel.uy, false);
      }
      if (thislevel.type == Quad.QUAD_CRITTER) {
        // this should be equivalent to thislevel.paint(graphics), but is
        // forceably inlined here.
        c   = (QuadCircleSpecies) thislevel;
        drawCircle(c.color, c.x, c.y, c.radius, false);
      } else if (!thislevel.isCollector(thislevel.type)) {
        thislevel.paint(this);
      }
    } while ((thislevel = thislevel.sibling) != null);
  }

  private int timey;
  private void drawTime() {
    if (!showLegend) return;
    long now = Globals.getCurrentTime();
    String snow = "Time " + now;
    int x = getSize().width - getStringWidth(snow);
    int y = getStringHeight();
    drawString(getForeground(), snow, x, y);
    timey = y;
  }

  private int getStringWidth(String s) {

    FontMetrics fm = getOffscreen().getFontMetrics();

    return fm.stringWidth(s) + 4;

  }

  private int getStringHeight() {

    return offg.getFontMetrics().getHeight() + 4;

  }

  /** returns 0 if only time is showing. */

  private int getLegendWidth() {

    if (!showLegend)

      return 0;

    int maxlen = getStringWidth("Time 00000");

    for (int i=0; i<toLegend.size(); i++) {

      int colorid = ((Integer)toLegend.elementAt(i)).intValue();

      String name = palette.getColorObj(colorid).name;

      if (getStringWidth(name) > maxlen)

        maxlen = getStringWidth(name);

    }

    return maxlen;

  }


  private void drawLegend() {
    if (!showLegend)
      return;

    for (int i=0; i<toLegend.size(); i++) {
      int colorid = ((Integer)toLegend.elementAt(i)).intValue();
      String name = palette.getName( colorid );
      int x = getSize().width - getStringWidth(name);
      int y = getStringHeight() * (1 + (i+1));
      drawString(palette.getColor( colorid ), name, x, y);
    }
  }


  synchronized public void paint(Graphics g) {
    amDrawing = true;
    if (curwin.isNullMcr()) {
      centerPicture(Globals.getWorldMcr());
    }

    // paint background on offscreen image (erase)
    Graphics offg = getOffscreen();
    synchronized (offg) {
      offg.setColor(getBackground());
      offg.fillRect(0,0, getCurrentSize().width, getCurrentSize().height);
    }
    // draw latest whole tree
    drawTime();
    drawTree();
    // put current time in upper right corner
    drawLegend();
    // super puts the offscreen, onscreen
    paintBetweenXorFlag = true; // communicate with zoombox
    super.paint(g);
    amDrawing = false;
  }

  /** center the world nicely on the screen, with margin for clarity */
  public void centerPicture(Mcr mcr) {
    if (mcr == null)
      return;

    getOffscreen();
    Dimension d = getCurrentSize();

    int margin = 20;
    d.width -= getLegendWidth();


    double xsize = mcr.ux - mcr.lx;

    double ysize = mcr.uy - mcr.ly;


    // fac = world per pix

    double facx = xsize / (double)(d.width - margin);
    double facy = ysize / (double)(d.height - margin);
    double fac = facx > facy ? facx : facy;

    // center world on window
    curwin.lx = 0.5 * ( (mcr.lx + mcr.ux) - (fac * d.width) );
    curwin.ux = curwin.lx + (fac * d.width);
    curwin.ly = 0.5 * ( (mcr.ly + mcr.uy) - (fac * d.height) );
    curwin.uy = curwin.ly + (fac * d.height);
    setPPW();
  }


  /** my zoom -- 3 in 1 a la Plots. Zoom out isn't very useful, so let's
   *  make a downbox or upbox do zoom in. Differs from Plot in that square
   *  aspect ratio is enforced, and working zoom is only zoom out. Simpler than
   *  Plot in that we don't have to screw with rounding axis ticks.
   */
  protected void zoom()
  {
    double new_xmin, new_ymin, new_xmax, new_ymax;
    int tolerance     = 10;		// moved < 10 pixels == didn't move.

    // tiny box == back to whole-scale
    if ((Math.abs(selection.width) < tolerance) ||
        (Math.abs(selection.height) < tolerance))
    {
      curwin = truewin;
      centerPicture(Globals.getWorldMcr());
      return;
    }
    if (selection.height < 0) {
      selection.y = selection.y + selection.height;
      selection.height = - selection.height;
    }
    if (selection.width < 0) {
      selection.x = selection.x + selection.width;
      selection.width = - selection.width;
    }

    new_ymin = p2wY(selection.y);
    new_ymax = p2wY(selection.y + selection.height);
    new_xmin = p2wX(selection.x);
    new_xmax = p2wX(selection.x + selection.width);
    curwin = new Mcr();
    curwin.setMcr(new_xmin, new_ymin, new_xmax, new_ymax);
    centerPicture(curwin);
  }

  public String toString() {
    return Globals.pointerString(this) + super.toString() + ",quadRoot=" + quadroot;
  }
}












