//Title:        BacSim/Gecko
//Version:
//Copyright:    Copyright (c) 1997-2001, Jan Kreft, All rights reserved
//Author:       Jan Kreft, Kreft@cardiff.ac.uk
//Company:      University of Cardiff
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Bacterial Ecology Simulator (BacSim)
/**
 * Bacillus is the basic agent of BacSim, living in a 3-d substrate
 * world implemented by Diffuse.
 * 990924, vfb, translated from the Swarm Objective-C BacSim source.
 * Note, Bacillus *should* extend QuadCircleSpecies instead of SizedPlant.
 * But need careful redesign to replace minsize, breedsize, etc. Think I
 * should just get it working first.... But it's rather high in the heirarchy:
 * QuadCircleSpecies -> Critter -> PassiveCritter -> SizedPlant -> Bacillus.
 * Jan and I went opposite directions in what we considered Critter-essential.
 *
 * 001209, vfb, update per Jan's objc changes.
 */
package gecko.bacsim;
import gecko.*;
import gecko.zoo.*;
import stats.*;
import gecko.statglue.BiomassTransfers;
import java.util.*;
import java.io.*;
import courseware.util.PrintfFormat;

public class Bacillus extends SizedPlant implements HasProcess {
  // this isn't used for anything, but assures that class has been loaded.
  // the constants weren't ever referred to?
  private static BacsimResources res = new BacsimResources();
  // global parameters valid for all Bacilli, set by protocol!
  private static double density;                   // the density of the cells (dryMass)
  private static double shoveRadius;		// multiplier on radius for shoving purposes
  private static double convertBack;		// when die, multiplier on mass->substrate
  private static boolean someHaveEps = false; // do we have eps at all?

  // vfb 010114, all per-species params moved to BacillusParam, where they have
  // package-level visibility, i.e., it's more a record than a class
  private BacillusParam params;

  // working vars
  // all these instance vars are variables
  private double vmax; // vfb, 010114, was listed as input param, but I dunno...varies in .ml o/p
  private double  myS1, myS2; // my subConcns
  private double  uptakeRate;
  private double dryMass, lastDryMass, growthRate;
  private double  epsMass;
  private double  epsRadius;
  private int  myGrandma; // which Grandma do I stem from
  // identifier for cell and lineage, my grandma 0, first division gives 00 and 01, etc.
  private String myName;

  // This property list is used to generate paramLists for popup editing.
  // Put them in whatever order you want them.
  private static String[] props = { "density", "shoveRadius", "convertBack",
    "breedSize", "process", "sub1", "sub2", "y1", "y2", "maintenanceR",
    "vmax", "km1", "km2", "ki1",
    "eps", "epsProdCoeff", "epsProdConst", "epsDensity" };
  public String[] getPropertyNames() { return props;}

  // static Bacillus vars--shared by all Bacillus
  static int grandmaCount = 0;
  private static Diffuse actualSite;
  private static LandRoot landRoot;
  private static Quad mobileRoot;
  private static Mcr landMcr;

  // info we get from DiffuseQuad, set via Protocol
  // let's stop copying information and confusing me....

  // these apparently had no hook to set them...
  private static boolean precise = false;
  private static boolean randomGrandma  = false;
  // public because most shoving is done by BacShove.
  public static boolean mutualShove = true;
  public static boolean squareShove = false;

  // stats for yields, per species and per substrate
  private static double[][] popFed; // popFed[species][sub]

  // avoid allocating this thing over and over
  private static McBrick shovewin = new McBrick();

  public Bacillus() {
    super();
    params = new BacillusParam();
    setPutOnPlantRoot(false); // go on mobileRoot
    myName = new String("0");  // for grandma of all grandmas
  }
  public Bacillus(BacillusParam bp) {
    super();
    this.params = bp;
    setPutOnPlantRoot(false);
    myName = new String("0");  // for grandma of all grandmas
  }
  public void setFrom(Bacillus other) {
    super.setFrom(other);
    params = other.params;
    vmax = other.vmax;
    // these are needed before runInitChecks--which reactions to init, I guess

    // someplace to put once-per-run initializations
    runInitChecks();

    // private but heritable?
    myName = new String(other.myName);
    myGrandma = other.myGrandma;

    // someplace to put species birth into world initializations
    // need k's first
    // may override grandma
    grandmaInitChecks(true);

    // non-heritable working aspects
    dryMass = density * Globals.radius2volume(radius);
    lastDryMass = dryMass;
    epsMass = 0.0; // EPS stuff only ever changed if eps flag is TRUE
    epsRadius = 0.0;
    update();

    z = actualSite.getLengthZ() * Globals.getRandomSource().nextDouble();
    setMcr();   // set mine
    growMcr();  // stretch quads above me if needed (normally isn't needed)
  }


  static private boolean makingFromFile = false;
  static public void makeAgentsFromFile(BufferedReader file) {
    makingFromFile = true;

    int i, j;
    int tokenCount = 0;
    int thisProcess;
    int lineNo = 0;
    Bacillus thisOne = null;
    StringTokenizer parser;
    int num = 0;
    int numread;
    int got; // result from FileReader.read()
    String line = null; // line read in
    String token;
    mobileRoot = Globals.getMobileRoot();
    landRoot = Globals.getLandRoot();
    actualSite = findSite();

    // vfb, 010114, used / timeunit on output, use * timeunit on input?
    int timeunit = actualSite.getStepSize();
    int now = (int) Globals.getCurrentTime();

    // line is null at eof
    try {
      while ( (line = file.readLine()) != null) {
        lineNo++;
        tokenCount = 0;
        parser = new StringTokenizer(line); // the default delimiters are " \t\n\r"
        token = parser.nextToken();
        tokenCount ++;

        thisProcess = Integer.valueOf(token).intValue();
        thisOne = (Bacillus) BacGlobals.findProgenitor(thisProcess).clone();
        thisOne.getSpecies().setBirthday(-1); // don't let species hang them on tree...
        if (thisOne == null) {
          Globals.log("ERROR: unable to find progenitor for process " + thisProcess);
          Globals.log("Line " + lineNo + " input file says:");
          Globals.log(line);
          Globals.getSimulator().crashStop();
        }

        while ( parser.hasMoreTokens() ) {
          token = parser.nextToken();
          tokenCount++;
          switch (tokenCount) {
            case 2:
              thisOne.x = Double.valueOf(token).doubleValue(); break;
            case 3:
              thisOne.y = Double.valueOf(token).doubleValue(); break;
            case 4:
              thisOne.z = Double.valueOf(token).doubleValue(); break;
            case 5:
              thisOne.radius = Double.valueOf(token).doubleValue();
              thisOne.dryMass = density * Globals.radius2volume( thisOne.radius );
              break;
            case 6:
              thisOne.myGrandma = Integer.valueOf(token).intValue(); break;
            case 7:
              thisOne.myName = new String(token); break;
            case 8:
              thisOne.growthRate = Double.valueOf(token).doubleValue();
              // vfb, 010114, used / timeunit on output, use * timeunit on input?
              // but not for .in files....
              if (now > 0) thisOne.growthRate *= timeunit;
              thisOne.lastDryMass = thisOne.dryMass - thisOne.growthRate;
              break;
            case 9:
              thisOne.vmax = Double.valueOf(token).doubleValue();
              // vfb, 010114, used / timeunit on output, use * timeunit on input?
              // but not for .in files....
              if (now > 0) thisOne.vmax *= timeunit;
              break;
            case 10:
              thisOne.epsRadius = Double.valueOf(token).doubleValue();
              if (thisOne.epsRadius > 0)
                thisOne.epsMass = thisOne.params.epsDensity *
                  (Globals.radius2volume(thisOne.epsRadius) - (thisOne.dryMass/density));
              else
                thisOne.epsMass = 0;
              break;
            default:
              Globals.log("Warning, unknown case reading input file, line " + lineNo + " reads:");
              Globals.log(line);
              break;
          } /*switch*/
        } /*while token*/

        // to be done for every agent/line-read-in
        thisOne.birthday 	= Globals.getCurrentTime();
        thisOne.kidcount 	= 0;
        thisOne.setMcr(); // after this, we have an mcr and can ask for our site
        thisOne.update();
        thisOne.grandmaInitChecks(false);
        mobileRoot.addChild(thisOne);
        thisOne.getSpecies().notifyBirth();
      } /*while line*/

      // vfb 010111, because centerpicture doesn't do too well....
      landRoot.setFrom(actualSite);

      // to be done once at the end
      if (thisOne != null)
        thisOne.runInitChecks();
      file.close();
    } catch (Exception e) {
      Globals.log("Unexpected error parsing input: " + e);
      if (line != null)
        Globals.log("Input line " + lineNo + " : " + line);
      else
        Globals.log("<input line null>, line " + lineNo);
      Globals.getSimulator().crashStop();
    }
    makingFromFile = false;
  }

  public Object clone() {return makeKid();}

  public Speciesable makeKid() {
    Bacillus kid = new Bacillus(params);
    kid.setFrom(this);
    //getSpecies().notifyBirthToMom(this); // for stages???
    return (Speciesable) kid;
  }

  static private Diffuse findSite() {
    // translation : take any diffuse quad we can find.
    Quad q = landRoot.pickWin(landRoot, landRoot, false);
    while (q!=null) {
      if (q instanceof Diffuse)
        return (Diffuse) q;
    }
    System.err.println("***Error, no Diffuse site found!");
    Globals.getSimulator().crashStop();
    return null; // returns nothing after a crashStop...
  }

  /** Place to stuff initialization. Note that in order to
   *  reset global vars to 0 on subsequent runs, about the only
   *  hook we've got is Species.holdData. If it's null, this is
   *  the first time this routine has been called since Species
   *  was told to reset. However, the holdData as of this writing
   *  has nothing -in- it--contents unnecessary, existence needed.
   *
   *  Things that happen once per program initialization use
   *  static initializers in the declarations section.
   */
  private void runInitChecks() {
    if (getSpecies().isNewRun()) {
      grandmaCount = 0;

      setMcr();    // need an Mcr before we can go picking with it.
      landRoot   = Globals.getLandRoot();
      actualSite = findSite();
      mobileRoot = Globals.getMobileRoot();

      // don't bother to reproduce this logic, just tell Eps....
      Eps.setSiteInfo(actualSite);


      popFed = new double[actualSite.getNumSpecies()][actualSite.getNumSubs()];
    }
  }


  /** Kludgy place to put "end of species.step hanging an instance of grandma"
   *  Do NOT use setBirthday during the run! Bacillus never calls this.
   */

  public void setBirthday(int i) {
    super.setBirthday(i);

    // OK, now my kludgefest
    double tempR;

    if (randomGrandma) { // random grandma switch
      // random size for grandma, but must be > minsize and < breedvol
      // don't change starting point radius
      do {
        tempR = deviateFrom(radius, 0.4);
      } while ( (tempR <= minsize) || ( tempR >= Globals.volume2radius(breedsize) ) );
    } else {
      tempR = radius;
    }

    growthRate = 0;

    // z goes after the update below
    if (precise) {
      // x, y already precise
      radius = tempR;
    } else {
      // reduce precision to 6 digits, same as in printing using %g
      String formatted = new PrintfFormat("%g").sprintf(x);
      x = Double.valueOf(formatted).doubleValue();
      formatted = new PrintfFormat("%g").sprintf(y);
      y = Double.valueOf(formatted).doubleValue();
      formatted = new PrintfFormat("%g").sprintf(radius);
      radius = Double.valueOf(formatted).doubleValue();
    }

    update(); // apparently screws with z, so now we can rescrew...

    if (precise) {
      z = actualSite.getLengthZ() * Globals.getRandomSource().nextDouble();
    } else {
      String formatted = new PrintfFormat("%g").sprintf(
          actualSite.getLengthZ() * Globals.getRandomSource().nextDouble());
      z = Double.valueOf(formatted).doubleValue();
    }

    checkBirthArea();
    BacShove.wrapXYZ(this,actualSite);

    setMcr(); // have to repeat after changing z
  }


  /** 001210: this would be an ideal place to put several kludgy things done by
   *  the objc "setXYR", except that JanK's latest changes use that routine
   *  to set the x, y values at random, and we can't do that here, because
   *  species.step calls setX, setY, setR -after- hanging the grandma clone.
   *  So what was in objc setX:Y:R is spread between this, a kludgy override
   *  on setBirthday, runInitChecks, maybe other places...
   */
  private void grandmaInitChecks(boolean setGrandma) {
    if ( (Globals.getCurrentTime() == getSpecies().getBirthday()) ||
         (makingFromFile ) )
    {
      if (setGrandma) {
        grandmaCount++;
        myGrandma = grandmaCount;
      }
      myName = new String("0");
      if (!params.convertedFlag) {
        // only sub1, the nitrogen substrate, has to be converted into neutral,
        // un-ionized form
        switch (params.sub1) {
        case 1: // ammonia
          params.km1 = params.km1 * Diffuse.KaNH3 / Math.pow(10,Diffuse.pH);
          params.ki1 = params.ki1 * Diffuse.KaNH3 / Math.pow(10,Diffuse.pH);
          break;
        case 2: // nitrite
          params.km1 = params.km1 * (Math.pow(10,Diffuse.pH) * Diffuse.KbHNO2);
          params.ki1 = params.ki1 * (Math.pow(10,Diffuse.pH) * Diffuse.KbHNO2);
          break;
        default: Globals.getUI().log("** error in Bacillus, sub1=" + params.sub1 + "\n");
        }

        // We normalize subConcn by dividing Kms and Kis and all subConcn
        // in DiffuseQuad by the external subConcn
        params.km1 /= actualSite.getExtConcn(params.sub1);
        params.ki1 /= actualSite.getExtConcn(params.sub1);
        params.km2 /= actualSite.getExtConcn(params.sub2);

        params.convertedFlag = true;
      }
    }
  }

  /** Note that breedsize has the meaning of Volume in contrast to
   * standard Gecko where all sizes are radii (see SizedPlant)
   *
   * translation: since setBreedSize only called from protocol, and
   * gecko now hangs protoagents at -radius- rather than -kidsize-,
   * let's make setBreedSize initialize radius to kidsize. Another
   * option would be to specify radius, which is more error prone.
   */
  public void setBreedSize(double bs) {
    double kidvol, minvol;
    breedsize = bs;
    kidvol   = 0.75 * breedsize;
    kidsize  = Globals.volume2radius(kidvol);
    minvol   = 0.2 * breedsize;
    minsize  = Globals.volume2radius(minvol);
    radius   = kidsize;
  }

  /** translation: operated on DParam before...no more.
   *  there were other deviates and varyFrom's--think this one the
   *  only one actually called.
   */
  private double deviateFrom(double value, double sigma) {
    double result;
    double deviation;
    double cv = sigma; // coefficient of variation = relative standard deviation
    do {
      // restrict change to a range of 2 * relative standard deviation
      do {
        deviation = Globals.getRandomSource().nextGaussian();
      } while (Math.abs(deviation) >= 2.0);
      result = value * (1.0 + cv * deviation);
    } while (!Globals.samesign(result, value));

    return result;
  }

  public double getRateForReaction(int r, double s1, double s2) {
    double rate = 0;//, s1n;

    switch (r) { // for which reaction should we supply the rate?
    // (We may have more than 4 cases in future)
    case 1:
      // Nitroso: ammonia consumption, s1 is ammonia and s2 is oxygen
      //s1n = s1 * pow(10,pH) / KaNH3; // We use the unionized form of the substrate
      rate = - dryMass * vmax * (s1/(params.km1+s1+(s1*s1/params.ki1))) * (s2/(params.km2+s2));
      break;
    case 2:
      // Nitroso: oxygen consumption, s1 is ammonia and s2 is oxygen
      //s1n = s1 * pow(10,pH) / KaNH3; // We use the unionized form of the substrate
      rate = - dryMass * (params.y1/params.y2) *
             vmax * (s1/(params.km1+s1+(s1*s1/params.ki1))) * (s2/(params.km2+s2));
      break;
    case 11:
      // Nitro: nitrite consumption, s1 is nitrite and s2 is oxygen
      //s1n = s1 / (pow(10,pH) * KbHNO2); // We use the unionized form of the substrate
      rate = - dryMass * vmax * (s1/(params.km1+s1+(s1*s1/params.ki1))) * (s2/(params.km2+s2));
      break;
    case 12:
      // Nitro: oxygen consumption, s1 is nitrite and s2 is oxygen
      //s1n = s1 / (pow(10,pH) * KbHNO2); // We use the unionized form of the substrate
      rate = - dryMass * (params.y1/params.y2) * vmax
             * (s1/(params.km1+s1+(s1*s1/params.ki1))) * (s2/(params.km2+s2));
      break;
    }
    return rate;
  }

  public void grow() {
    // inquire about all relevant subConcn's
    myS1 = actualSite.avgSub(params.sub1, x, y, radius);
    myS2 = actualSite.avgSub(params.sub2, x, y, radius);

    // grow: take up sub, metabolize sub taken up completely, and maintain
    // growth independent substrate consumption = maintenance rate
    // maintenance is debited from dryMass, hence it's independent of substrates

    // The meaning of rate is amount per timestep, as the timestep is implicit,
    // rates are actually amounts
    uptakeRate = dryMass * vmax *
                (myS1/(params.km1+myS1+(myS1*myS1/params.ki1)))
                * (myS2/(params.km2+myS2));

    popFed[params.process-1][params.sub1]  += uptakeRate;
    popFed[params.process-1][params.sub2]  += uptakeRate * (params.y1/params.y2);

    // this is not specific growth rate but dx/dt
    // also, this is not per min, but per timestep, in printv it is per min
    if (params.eps == Eps.NOEPS) {
      growthRate = (uptakeRate * params.y1) - (dryMass * params.maintenanceR);
    }
    else {
    // EPS production is a function of growth a la Luedeking-Piret expression
      growthRate = (uptakeRate * params.y1 - dryMass *
                    (params.maintenanceR + 1.214286*params.epsProdConst)) /
                    (1 + 1.214286*params.epsProdCoeff);
      epsMass += params.epsProdCoeff * growthRate + dryMass * params.epsProdConst;
      Eps.notifyBiomass(epsMass);
    }

    lastDryMass = dryMass;
    dryMass += growthRate;
  }

  // dryMass and lastDryMass are set outside and before update
  // update only deals with dependent variables
  private void update() {
    double distance, xi, yi, zi, tmp;
    double alpha, beta;
    long now = Globals.getCurrentTime();

    radius = Globals.volume2radius(dryMass / density);
    setMcr();

    switch (params.eps) {
      default: break;
      case Eps.CAPSULE:
        epsRadius = Globals.volume2radius( (dryMass / density) + (epsMass / params.epsDensity) );
        break;
      case Eps.SLIME: case Eps.SELFBINDING: case Eps.ALLBINDING: case Eps.SLIMERSBINDING:

        if ((epsMass / params.epsDensity) > breedsize) { // transfer epsMass from this Bacillus to Eps

          // We want eps blob and excreter adjacent with the same center (volume not mass) as excreter before excretion,
          // both placed in a random direction
          distance = Math.pow( ((dryMass / density) + (epsMass / params.epsDensity)) , 1.0/3.0 ) / 2.0;
          Random randy = Globals.getRandomSource();
          alpha = 2.0 * Math.PI * randy.nextDouble();
          beta  = 2.0 * Math.PI * randy.nextDouble();
          // shift eps blob and self
          tmp = distance * Math.cos(alpha) * Math.cos(beta);
          xi  = x + tmp;
          x  -= tmp;
          tmp = distance * Math.sin(alpha) * Math.cos(beta);
          yi  = y + tmp;
          y  -= tmp;
          tmp = distance * Math.sin(beta);
          zi  = z + tmp;
          z  -= tmp;

          // Out of bounds check for Bacillus is done here, and for Eps is done in excreteEpsAmount
          BacShove.wrapXYZ(this,actualSite);

          Eps.excreteEpsAmount(epsMass, xi, yi, zi);
          epsMass = 0.0;
          epsRadius = radius;
        } else {
          epsRadius = Globals.volume2radius( (dryMass / density) + (epsMass / params.epsDensity) );
        }
    }
    //getSpecies().notifyBioTrans(BiomassTransfers.GROWTH, growthRate);
  }

  private boolean checkDeath() {
    if (radius < minsize) {
      die();
      return true;
    }
    return false;
  }

  public void die() {
    // see LandRoot, in future better not let LandRoot deal with this, not enuf control over which substrate where will get upped
    Globals.getTaxCollector().payTax(x, y, 0, dryMass * convertBack);

    Eps.excreteEpsAmount(epsMass, x, y, z);
    remove();

    Species s = getSpecies();
    s.notifyDeath();  // for population
    //s.notifyDeath(this); // for stage
    //s.notifyBioTrans(BiomassTransfers.LOSS_STARVE, dryMass);

    Globals.log("Death of " + params.process + " at " + Globals.getCurrentTime());
  }

  // descriptive Donachie model
  private void checkVolume() {
    double breedVol;
    double gph; // generations per hour

    gph = growthRate * 60 / (dryMass * actualSite.getStepSize() * Math.log(2));
    if (gph > 0) {
      // breedsize in Protocol is (Vu of Donachie = 0.4 fl) * (some factor):
      // see lab book page 147
      breedVol = breedsize * Math.pow(2.0,gph);

      // if want to make vol measurements imprecise
      // translation : was varyDValue--if diff, I can't see it.
      breedVol = deviateFrom(breedVol, 0.1);

      if ((dryMass / density) >= breedVol)
        reproduce();
    }
  }

  public boolean reproduce() {
    // False if didn't.

    Bacillus baby;
    double distance;
    double alpha, beta;
    Random randy = Globals.getRandomSource();
    long now = Globals.getCurrentTime();

    // mutate params here if we don't want diverse grandmas
    // but changes once every cell cycle, baby and mom equal
    // vary
    //vmax = [self varyDValue: vmax bySigma: 0.1];
    // deviate
    // translation: this doesn't change in the course of a run....
    // could get once in runInitChecks
    double origVmax = ((Bacillus) (getSpecies().getProgenitor()) ).vmax;
    vmax = deviateFrom(origVmax, 0.1) * actualSite.getStepSize();
    // make the kid
    Speciesable lbaby = makeKid();
    baby = (Bacillus) lbaby;

    // life stats, kidcount and birthday are inherited from Critter, BUT
    // kidcount is used as a generation counter here and
    // birthday has the meaning of time of last division
    // kidcount and birthday for mom = kid, since binary fission is:
    // mum splits into two daughters, no mum left
    baby.kidcount = kidcount = kidcount + 1;
    baby.birthday = birthday = Globals.getCurrentTime();
    myName.concat("0"); // lineage string is expanded one digit
    baby.myName.concat("1");

    // update variables, don't call update (update should only be called once per step)
    baby.dryMass = dryMass = dryMass / 2.0;
    baby.lastDryMass = lastDryMass = lastDryMass / 2.0;
    baby.radius = radius = Globals.volume2radius(dryMass / density);
    if (params.eps!=Eps.NOEPS) {
      epsMass /= 2.0;
      // epsRadius = Math.pow( ((dryMass / density) + (epsMass / epsDensity)) , 1.0/3.0 ) / 2.0;
      epsRadius = Globals.volume2radius( (dryMass / density) + (epsMass / params.epsDensity) );
      baby.epsMass = epsMass;
      baby.epsRadius = epsRadius;
    }

    // We want kid and mum directly adjacent with the same center as mum before reproduce,
    // both placed in a random direction
    distance = radius;
    if ((2*radius) > actualSite.getLengthZ()) {
      Globals.log("Warning, for " + params.process + " at " + now
        + " diam=" + 2*radius + " > sliceThickness");
    }
    alpha = 2.0 * Math.PI * randy.nextDouble();
    beta  = 2.0 * Math.PI * randy.nextDouble();
    // shift kid
    baby.x      = x + ( distance * Math.cos(alpha) * Math.cos(beta) );
    baby.y      = y + ( distance * Math.sin(alpha) * Math.cos(beta) );
    baby.z      = z + ( distance * Math.sin(beta)                   );
    // now I move, too....
    x -= ( distance * Math.cos(alpha) * Math.cos(beta) );
    y -= ( distance * Math.sin(alpha) * Math.cos(beta) );
    z -= ( distance * Math.sin(beta)                   );

    BacShove.wrapXYZ(this,actualSite);
    BacShove.wrapXYZ(baby,actualSite);

    // hang on quad tree
    baby.setMcr();
    // vfb, 991123, addNewbie pushes baby into quadtree instead of adding
    // at the front--somewhat better performance between treesorts.
    mobileRoot.addNewbie(baby);
    getSpecies().notifyBirth();
    getSpecies().notifyBiomass(baby.dryMass);

    growMcr();

    return true;
  }

  private void shove() {
    double shoveR;
    double overlapRadius = 0;
    double distRadius = 0;
    Quad q;
    QuadCircleSpecies neighbor;
    Direction hisDirection;

    if (!mutualShove)
      direction.zeroDirection();

    // initialize shove window
    // if we are producing eps, our distance to neighbors is explicitly
    // modelled by the total volume taken up by
    // EPS and biomass, else we multiply our radius by a calibrated factor
    // to avoid overlap between cells
    switch (params.eps) {
      case Eps.NOEPS:
        distRadius = shoveRadius * radius;
        shovewin.sphereToBrick(x, y, z, distRadius);
        break;
      case Eps.CAPSULE:
        distRadius = shoveRadius * epsRadius;
        shovewin.sphereToBrick(x, y, z, distRadius);
        break;
      case Eps.SLIME:
        distRadius = shoveRadius * epsRadius;
        shovewin.sphereToBrick(x, y, z, distRadius);
        break;
      default :
        distRadius = shoveRadius * epsRadius;
        // need larger shovewin if we also deal with attractive forces
        shovewin.sphereToBrick(x, y, z, 3.0*distRadius);
        break;
      }

    // 1. Shove remainder of quadding tree starting from me. Note that my
    // direction already contains shoves from every neighbor before me in the
    // tree, done by this routine for them.

    q = this.pickWin(shovewin, mobileRoot, false);
    while (q != null) {
      // do one mutual shove
      neighbor = (QuadCircleSpecies) q;
      overlapRadius = distRadius;
      overlapRadius -= Globals.distance( x, y, z,
          neighbor.x, neighbor.y, neighbor.z ) - neighbor.radius;
      if (overlapRadius > 0) {
        // my side of the shove
        if (squareShove)
          overlapRadius *= overlapRadius;

        direction.addDirection(x - neighbor.x,
                     y - neighbor.y,
                     z - neighbor.z, overlapRadius);

        // his side of the shove
        if (mutualShove) {
          hisDirection = ((GetDirectionable)neighbor).getDirection();

          hisDirection.addDirection(neighbor.x - x,
                       neighbor.y - y,
                       neighbor.z - z, overlapRadius);
        }
      } else {
        // cells too far? note we are not eps, but some eps attract...
        // attractive force has to be negative, weak, slowly rising
        // at the beginning, and fading in the distance
        double weight = -0.01 * (overlapRadius*overlapRadius) /
                ( (-overlapRadius+1) * (-overlapRadius+1) * (-overlapRadius+1) );
        switch (params.eps) {
          default: break; // do nothing
          case Eps.SELFBINDING: // neighbor same species? Note that we are not eps.
            if (params.process == ( ((Bacillus)neighbor).getProcess() ) )
              direction.addDirection(x - neighbor.x, y - neighbor.y, z - neighbor.z, weight);
            break;
          case Eps.ALLBINDING: // neighbor any species, that is, not eps?
            // vfb--eps process == -1
            if ( ((Bacillus)neighbor).getProcess() >= 0)
              direction.addDirection(x - neighbor.x, y - neighbor.y, z - neighbor.z, weight);
            break;
          case Eps.SLIMERSBINDING:
            if ( ((Bacillus)neighbor).getEps() != Eps.NOEPS)
              direction.addDirection(x - neighbor.x, y - neighbor.y, z - neighbor.z, weight);
            break;
        }
      }
      // next
      q = q.pickWin(shovewin, mobileRoot, false);
    }

    // 2. check the carrier wall
    if (actualSite.isBiofilm()) {
      // Then get a shove from the concrete wall called carrier as well
      // (none if we don't sit on carrier)
      // carrier is always at the low x side
      overlapRadius = actualSite.getOverlapWithCarrierAt((x - distRadius), y);
      if (overlapRadius > 0) {
        direction.addDirection(1, 0, 0, overlapRadius);
      }
    }

    // 3. apply all the shoves to date (neighbors and carrier)
    double u[] = direction.getWannas();
    x += u[0]; y += u[1]; z += u[2];

    // 4. wrap boundary conditions
    BacShove.wrapXYZ(this,actualSite);

    // 5. zero out direction--have applied all--and correct mcr
    if (mutualShove)
      direction.zeroDirection();
    setMcr();
  }

  public void checkBirthArea() {
    if (actualSite.getGeometry() == Diffuse.BIOFILM_GEOMETRY) {
      // cells should not be within the carrier nor in the topmost layer
      // which is always bulk liquid
      if (x <= actualSite.getCarrierLayer()) {
        x = actualSite.getCarrierLayer() + BacGlobals.EPSILON;
        Globals.log("Warning, x < cellSize, change birtharea for " + params.process + " in protocol.");
      }
      if (x >= (actualSite.getLengthX() - actualSite.getCellSize())) {
        x = actualSite.getLengthX() - actualSite.getCellSize() - BacGlobals.EPSILON;
        Globals.log("Warning, x > (lengthX-cellSize), change birtharea for " + params.process + " in protocol.");
      }
    }
  }

  public void step() {
    // One step in the lifecycle of a Bacillus.
    // Note: returns nil if dies--higher level routine deletes, because
    // higher level routine is traversing the tree.

    // print before everything else since most things are printed only when
    // it's the first cell's turn
    /* translation: hold off on printv's
    if ((Globals.getCurrentTime() % Globals.getStatCycle()) == 0) {
      printv();
    }
    */
    long now = Globals.getCurrentTime();
    if ((birthday == now) && (now > 0)) return; // we stepped already

    grow();
    update();
    // vfb, added 991123, cut down on doTheLocomotion
    shove();
    if (checkDeath()) // death if radius < minsize
      return;
    checkVolume(); // are we big enough to trigger division?

    getSpecies().notifyBiomass(dryMass);
    //getSpecies().notifyAlive(this); // for stages
  }

  // Diffuse wants to know
  public double getDryMass() {return dryMass;}
  public double getGrowthRate() {return growthRate;}

  // get/set for each entry in myprops ("isX" is standard for boolean "get")
  public boolean haveEps() {return someHaveEps;}
  public int getEps() {return params.eps;}
  public void setEps(int i){
    params.eps = i;
    if (i != Eps.NOEPS)
      someHaveEps = true;
  }
  public boolean isEps() {
    return (params.eps != Eps.NOEPS);
  }
  public int getProcess (){return params.process;}
  public void setProcess (int v){params.process = v;}
  public int getSub1 (){return params.sub1;}
  public void setSub1 (int v){params.sub1 = v;}
  public int getSub2 (){return params.sub2;}
  public void setSub2 (int v){params.sub2 = v;}
  public double getY1 (){return params.y1;}
  public void setY1 (double v){params.y1 = v;}
  public double getY2 (){return params.y2;}
  public void setY2 (double v){params.y2 = v;}
  public double getVmax (){return vmax;}
  public void setVmax (double v){vmax = v;}
  public double getKm1 (){return params.km1;}
  public void setKm1 (double v){params.km1 = v;}
  public double getKm2 (){return params.km2;}
  public void setKm2 (double v){params.km2 = v;}
  public double getKi1 (){return params.ki1;}
  public void setKi1 (double v){params.ki1 = v;}
  public double getMaintenanceR (){return params.maintenanceR;}
  public void setMaintenanceR (double v){params.maintenanceR = v;}
  public double getEpsProdCoeff (){return params.epsProdCoeff;}
  public void setEpsProdCoeff (double v){params.epsProdCoeff = v;}
  public double getEpsProdConst (){return params.epsProdConst;}
  public void setEpsProdConst (double v){params.epsProdConst = v;}
  public double getEpsDensity (){return params.epsDensity;}
  public void setEpsDensity (double v){params.epsDensity = v;}
  public double getDensity(){return density;}
  public void setDensity(double v){density = v;}
  public double getShoveRadius(){return shoveRadius;}
  public void setShoveRadius(double v){shoveRadius = v;}
  public double getConvertBack(){return convertBack;}
  public void setConvertBack(double v){convertBack = v;}
  public boolean putOnPlantRoot() {return false;}

  // these had no access hooks in objc...were inside a routine
  public void setPreciseXYZR(boolean b) {precise = b;}
  public void setRandomGrandma(boolean b) {randomGrandma = b;}
  public void setMutualShove(boolean b) {mutualShove = b;}
  static public void setSquareShove(boolean b) {squareShove = b;}
  static public boolean isSquareShove() {return squareShove;}

  public String toString() {
    return "Bacillus" + Globals.pointerString(this) +
            "[type=" + type + ",process=" + params.process + ",x=" + x + ",y=" + y + ",z=" + z + "]";
  }


  // BacShove wants to know.
  public double getEpsRadius() {return epsRadius;}

  /** Debug routine, not normally used. Skeleton, add what you want to look at,
   *  when and if.
   */
  private static int lastRound = -1;
  private static int lastRun = -1;
  public void printv() {
    // most printing is done from Diffuse while the growth step is frozen,
    // here we only print when we try things out
    // some things should be printed for all cells, some not
    // some things should be printed often, some not

    int now = (int)Globals.getCurrentTime();
    int thisRun = Globals.getGeckoDriver().getNumRun();

    // this is printed only for one cell per round
    if ( (lastRound != now) || (lastRun != thisRun) ) {
      // this is printed at lower frequency ([geckoControl getStatCycle])
      if ( ( now % Globals.getStatCycle() ) == 0 ) {

      }

      // this is printed every printstep

      lastRound = now;
      lastRun = thisRun;
    }

    // this is printed for all cells

    // this is printed at lower frequency ([geckoControl getStatCycle])
    if ( ( now % Globals.getStatCycle() ) == 0) {

    }
  }

  private static PrintWriter sphereFile;
  static public void closeSphereFile() {
    if (sphereFile ==  null)
      return;
    sphereFile.flush();
    sphereFile.close();
    BacGlobals.flushFiles();
  }


  static private void clearSpheresMatlab(boolean leaveOpen) {

    // overwrite--zero out the file

    int now = (int)Globals.getCurrentTime();
    int thisRun = Globals.getGeckoDriver().getNumRun();
    String filename = "<uninit>";
    try {
      // make sure last one closes politely
      if (sphereFile != null)
        sphereFile.close();

      filename = ((BatchUI) Globals.getUI()).getFileStem();
      filename += new PrintfFormat("_%03ld_").sprintf(thisRun);
      filename += new PrintfFormat("%06ld.ml").sprintf(now);
      // no append--clear it
      FileWriter filewriter = new FileWriter(filename, false);
      filewriter.close();
      if (leaveOpen) {
        // now open in append mode
        filewriter = new FileWriter(filename, true);
        sphereFile = new PrintWriter(filewriter, true);
      }
    } catch (Exception e) {
      Globals.log("Error : " + e + " Opening output file : " + filename);
      return;
    }
  }

  public void printSpheresMatlab() {
    // Replacement for snapshots, prints all agents for import into Matlab
    // or conversion into POV-Ray file
    int now = (int)Globals.getCurrentTime();
    int thisRun = Globals.getGeckoDriver().getNumRun();
    int timeunit = actualSite.getStepSize();

    // do this only once at the beginning of each round
    if ((lastRound != now) || (lastRun != thisRun)) {
      // overwrite--zero out the file, then open to append
      clearSpheresMatlab(true);
      lastRound = now;
      lastRun = thisRun;
    }

    // this time, append instead of overwrite
    try {
      sphereFile.print(new PrintfFormat("%d\t").sprintf(params.process)); // numerical identifier for species
      sphereFile.print(new PrintfFormat("%g\t").sprintf(x));  // x coordinate of center
      sphereFile.print(new PrintfFormat("%g\t").sprintf(y));  // y coordinate of center
      sphereFile.print(new PrintfFormat("%g\t").sprintf(z));  // z coordinate of center
      sphereFile.print(new PrintfFormat("%g\t").sprintf(radius));       // radius
      sphereFile.print(new PrintfFormat("%d\t").sprintf(myGrandma));    // number of grandma
      sphereFile.print(new PrintfFormat("%s\t").sprintf(myName));       // myName, string of 0 and 1
      sphereFile.print(new PrintfFormat("%g\t").sprintf(growthRate / timeunit));   // growth rate
      sphereFile.print(new PrintfFormat("%g\t").sprintf(vmax / timeunit));         // vmax
      sphereFile.print(new PrintfFormat("%g\n").sprintf(epsRadius));    // epsRadius
    } catch (Exception e) {
      Globals.log("Error : " + e + " writing ml output file at time " + now);
    }
  }

  public void printFed() {
    // print the amount of substrate taken up, per substrate and
    // species, useful for yields
    int i, j;
    PrintWriter bacFile = BacGlobals.getBacFile();
    if (bacFile==null)
      return;

    bacFile.print( new PrintfFormat( "%d\t" ).sprintf( Globals.getGeckoDriver().getNumRun() ) );
    bacFile.print( new PrintfFormat( "%d\t" ).sprintf( Globals.getCurrentTime() ) );

    for (i=0; i < actualSite.getNumSpecies(); i++)
      for (j=0; j < actualSite.getNumSubs(); j++)
        bacFile.print( new PrintfFormat( "%g\t" ).sprintf( popFed[i][j] ) );

    bacFile.println();
  }



} /* Bacillus */


