//Title:        Gecko Woods
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  "mini-Sortie"
/**
 * 991207, vfb, original Java
 * ReaperDeer happens every year, with the deer level between min and max
 * deer damage, and affecting all tree crowns below deerHeight. (2m default.)
 * Basically the deer eat the bottom of the tree crowns, inflicting between
 * min and max damage per year (varies by year, not tree--tree damage variation
 * comes only from how low their crowns grow.) Calls Tree.crownDamageBottom.
 *
 * ReaperDeer was originally envisioned as an edge effect, like Wind, but...
 * since deer can access anyplace in a hectare from the edge, it's probably
 * best to say they're here or they're not, because the real test is a
 * pathway in, and that's beyond what I want to do here....
 */

package gecko.woods;
import gecko.*;
import courseware.jfcdep.TruncVal;

public class ReaperDeer implements Reapable, Steppable {
  // defining
  double minDamage, maxDamage;
  double deerHeight = 2.0; // meters, deer reach

  // working
  private double thisYearDamage;

  public ReaperDeer() {
    minDamage=maxDamage=0;
    thisYearDamage = 0;
  }
  private int damageCount;
  public void step() {
    thisYearDamage = minDamage + Globals.getRandomDouble() * (maxDamage-minDamage);
    if ((thisYearDamage > 0) && (Globals.getCurrentTime() % Globals.getStatCycle()) == 0) {
      Globals.log("Deer damaged " + damageCount +
        " trees last year; for time " + Globals.getCurrentTime() +
        " inflict " + TruncVal.stringVal(thisYearDamage, 5));
      }
    damageCount = 0;
  }
  public void reap(ReapableTree tree) {
    if (thisYearDamage == 0)
      return;
    // don't munch 'em until they get to grow....
    if (tree.getAge() < 5)
      return;
    if (tree.getCrownBottom() < deerHeight)
      damageCount++;
    tree.crownDamageBottom(thisYearDamage, deerHeight);
  }
  public void setMinDamage(double d) {minDamage = d;}
  public double getMinDamage() {return minDamage;}
  public void setMaxDamage(double d) {maxDamage = d;}
  public double getMaxDamage() {return maxDamage;}
  public void setDeerHeight(double d) {deerHeight = d;}
  public double getDeerHeight() {return deerHeight;}
}
