//Title:        Gecko Woods
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  "mini-Sortie"
/**
 * 991208, vfb, original Java
 * ReaperOld applies a mortality probability each year only to trees which
 * say they're full grown. Senescence, basically. At present, I don't think
 * there's much point to a ReaperYoung.
 *
 * There are two modes here: either apply a fixed deathrate to all fullgrown
 * trees, or use the species mortRandom to figure a maxAge on the tree, and
 * find a deathrate to achieve that.
 */


package gecko.woods;
import gecko.*;

public class ReaperOld implements Reapable, Steppable {
  double deathrate = 0;
  boolean useSpeciesMort = false;
  public ReaperOld() {
    deathrate = 0;
    useSpeciesMort = false;
  }
  public void step() {  }
  public void reap(ReapableTree tree) {
    if ((!useSpeciesMort) && (deathrate == 0)) return;
    if (tree.isFullGrown()) {
      double death;
      if (useSpeciesMort) {
        // we don't want to use MortRandom directly here, but
        // to discount it so 1/mortRandom = maxAge. Our fullgrown
        // tree has already lived awhile.
        double maxage = 1.0/tree.getMortRandom();
        if (Double.isNaN(maxage))
          return;
        // chance of dying today is 1.0 over years left to max
        if (tree.getAge() > maxage)
          death = 1.0;
        else
          death = 1.0 / (maxage - tree.getAge());
      } else {
        death = deathrate;
      }
      if (Globals.getRandomDouble() < death) {
        tree.die();
      }
    }
  }
  public void setMort(double d) {deathrate = d;}
  public double getMort() {return deathrate;}
  public boolean isUseSpeciesMort(){return useSpeciesMort;}
  public void setUseSpeciesMort(boolean v) {useSpeciesMort = v;}

}
