//Title:        Gecko Woods
//Version:
//Copyright:    Copyright (c) 2000, All rights reserved
//Author:       Ginger Booth, ginger@gingerbooth.com
//Company:      Yale IBS/CCE
//Description:  "mini-Sortie"
/**
 * 000727: set up my partition-based ("coral model") stab at new model.
 * Basic idea is that only leafy crown and this-years-stem/coarse roots layer
 * and the fine roots are alive. Old wood is like a coral reef. And this
 * year's growth has to pay for fine root and leaf turnover.
 *
 * 000718: set up Tree2 as testbed, quit changing Tree for further
 * Sylvia-preparation explorations.
 */

package gecko.woods;
import gecko.*;
import gecko.zoo.*;
import gecko.statglue.*;
import courseware.util.*;
import courseware.plot.*;
import courseware.jfcdep.*;
import java.awt.*;
import java.util.Random;


public class Tree2 extends Tree {

  public Tree2() {
    this(new Tree2Params());
  }
  public Tree2(Tree2Params tp) {
    super(tp);
  }

  public Object clone() {
    return makeKid();
  }

  public Speciesable makeKid() {

    Tree2 kid = new Tree2((Tree2Params) params);

    kid.setBirthday(Globals.getCurrentTime());
    kid.setFrom(this);

    kid.fullGrownFlag = false;

    kid.sync0();

    getSpecies().notifyBirthToMom(this);
    return (Speciesable) kid;
  }

  /** Set all the derived sizes based on stemDiameter.
   *  CAREFUL: since radius is incremental, cannot call syncWithStem twice
   *  in one step, or the crown growth multiplies.
* GINGER: REMEMBER haven't recast nutrient factor to new scheme yet.
   * will need to change grow() when deal with that, and also, may need
   * to re-arrange report of BIOMASS_GROWTH in grow().
   *
   * So far, only change in grow() was done in super--check Woods.getGLI.
   */
  protected void syncWithStem() {
    double oldCrownDepth = z - crownBottom;
    Tree2Params params2 = (Tree2Params)params;

    // not a lot we can do about it--agent shoved offsite, has no area,
    // has no nutrientFactor--kill it.
    if (nutrientFactor == 0) {
      radius = 0;
      return;
    }

    // Pacala et al '93, eqn 1.
    // vfb, complicated 991207: deer damage crops crown from the bottom, so
    // restate so that deer damage isn't repeated beyond reason. If nothing
    // raises crownBottom (like deer does), is equivalent to old statement.
    double effCrownDepthToHeight = (z - crownBottom)/z;
    double stuntedMax = nutrientFactor * params.maxHeight;
    if (stuntedMax > params.maxHeight)
      stuntedMax = params.maxHeight;
if (Double.isNaN(stemDiameter)){
  System.err.println("stemDiameter is NaN, tree=" + this);
  Simulator.crashStop();
}

    z = stuntedMax *
        ( 1 - Math.exp( - stemDiameter * (params.heightInitSlope/params.maxHeight) ) );
    if (z > Woods.getMaxHeight())
      Woods.setMaxHeight(z);
    if (z >= stuntedMax)
      fullGrownFlag = true;

    crownBottom = z * (1 - effCrownDepthToHeight);
    uncroppedCrownBottom = z * (1 - params.crownDepthToHeight);

    // new model: pay the piper and find new crown radius
    double stemR = stemDiameter/2.0;

    // our "fresh wood" is our handle on this year's biomass budget--the
    // partition schedule says how much is in the budget for the other costs
    // of living/growth. Say stemBranch is a hollow cylinder going halfway
    // up the crown.
    double crownDepth = z - crownBottom;
    double deltaBA = Math.PI * ( (stemR * stemR) - (lastR * lastR) );
    double freshWood = params2.getFreshWood(z, crownDepth, deltaBA);
    if (Double.isNaN(freshWood)) { // then freshWood is NaN now...
      if (getAge() == 0)
        freshWood = params2.getFreshWood(z, crownDepth, Math.PI*stemR*stemR);
      else
        freshWood = 0;
    }

    double growthBudget = freshWood / params2.stemPart;
    double leafBudget = growthBudget * params2.leafPart;
    // added 000811, helps on edge trees, not sitequal < 1.0 (ba quadruples)
    leafBudget *= nutrientFactor;
    // note that leafBudget is growth at the density of wood. Turn it into
    // leaf volume.
    double leafVolumeWas = params2.coneVolume(radius, oldCrownDepth);
    double leafMassWas = leafVolumeWas * params2.getCrownDensity();
    double leafLoss = leafMassWas * params2.leafTurnover;
    leafMassWas -= leafLoss;
    leafMassWas += leafBudget;
    leafVolumeWas = leafMassWas / params2.getCrownDensity();
    double lastcrown = radius;
    radius = params2.coneRadius(leafVolumeWas, crownDepth);
    double radiusGrew = radius / lastcrown;
    uncroppedRadius *= radiusGrew;
if (Double.isNaN(radius)){
  System.err.println("radius is NaN, tree=" + this);
  Simulator.crashStop();
}
if (Double.isNaN(radiusGrew)){
  System.err.println("radiusGrew is NaN, tree=" + this);
  Simulator.crashStop();
}

if ( ( (Globals.getCurrentTime() % 5) == 0) &&
     (getBirthday() == 0) && (z > .75 * Woods.getMaxHeight() ) )
System.err.println("t" + Globals.getCurrentTime() + ", stemD=" + stemDiameter + ",h=" + z);

//if (z > 0.95*params2.maxHeight) {
//System.err.println("t" + Globals.getCurrentTime() + ", z%=" + (z/params2.maxHeight) +
//", crownR=" + radius + ", leafBudget=" + leafBudget
//+ ", crownDepth=" + crownDepth + ", deltaBA=" + deltaBA);
//}


// GINGER: let's try this without cropping to maxCrownRadius first, and
// see what happens.

// System.err.println("time " + Globals.getCurrentTime() +
// ", height=" + z + ", stemD=" + stemDiameter + ", radius = " + radius);




// find maxCrown at 95% maxheight given Sortie's parameters
// also turned reproduction off so didn't get crowded.
/*
if ( (z > (.94 * params.maxHeight)) && (z < (.96 * params.maxHeight)) ) {
System.err.println("time " + Globals.getCurrentTime() + " : " +
getSpecies().getName() + " height = " + TruncVal.truncVal(z/params.maxHeight)
+ " of max, trunkD=" + stemDiameter + ", crownR = " + TruncVal.truncVal(radius));
}
*/
// if ((Globals.getCurrentTime() % 10) == 0)
// System.err.println("crown = " + radius + ", z=" + z);

    setMcr();

  }

  /** That part of syncWithStem that makeKid needs to see to. */
  protected void sync0() {
    Tree2Params params2 = (Tree2Params)params;

    z = params2.maxHeight *
        ( 1.0 - Math.exp(
          - (params2.recruitRadius * 2) * (params.heightInitSlope/params.maxHeight) ) );
    if (z > Woods.getMaxHeight())
      Woods.setMaxHeight(z);
    crownBottom =
    uncroppedCrownBottom = z * (1 - params.crownDepthToHeight);

    // new model: pay the piper and find new crown radius
    stemDiameter = params2.recruitRadius * 2.0;
    uncroppedRadius =
    radius = stemDiameter * params2.crownToStem;

    setMcr();
  }

  protected static String[] props = { "recruitRadius", "maxHeight", "heightInitSlope",
    "crownToStem", "crownDepthToHeight", "maxCrownRadius", "lightCoeff", "growRateMax",
    "growRateMin", "growRingMax",
    "recruitMeanD", "bigFertility",
    "stemPart", "leafPart", "leafTurnover",
    "mortRandom"};
  public String[] getPropertyNames() {return props;}

  public void setStemPart(double value) {
    ((Tree2Params)params).stemPart = value;
  }
  public void setLeafPart(double value) {
    ((Tree2Params)params).leafPart = value;
  }
  public void setLeafTurnover(double annual) {
    ((Tree2Params)params).leafTurnover = annual;
  }
  public double getStemPart(){return ((Tree2Params)params).stemPart;}
  public double getLeafPart(){return ((Tree2Params)params).leafPart;}
  public double getLeafTurnover(){return ((Tree2Params)params).leafTurnover;}

  /** Experiment, 000802, try debiting both trees, using crownDensity instead
   *  of lightcoeff.  Dreadful. Try mutual half?
   */
  protected void crownCrop(Lightbox lightbox, McrCircle equiv) {
    // set up for trunk aversion checks
    // my trunk-immune-to-seedlings area, applied in bask().
    double stemAversionR = Woods.thinningFactor * stemDiameter / 2.0; // that's cm
    stemAversionR /= 100; // that's m
    stem.circleToSquare(x, y, stemAversionR);

    // equiv and wannago preset by bask
    Quad plantRoot = Globals.getPlantRoot();
    Quad somequad = plantRoot.pickWin(equiv, plantRoot, false);
    Tree shadower;
    McBrick myBrick = new McBrick(this, crownBottom, z);
    McBrick crossBrick;
    double crownVolume = myBrick.getVolume();
    double crossVolume;
    while (somequad != null) {
      if ((somequad instanceof Tree) && (somequad != this)) {
        // first issue--crown aversion
        shadower = (Tree) somequad;
        if (Woods.isCrownCrop()) {
          // Does shadower affect me? already know our mcr's intersect, but not volume
// make it mutual
//          if (shadower.z > z) {
//            if (shadower.crownBottom < z) {
              crossBrick = new McBrick(shadower, shadower.crownBottom, shadower.z);
              if ( myBrick.intersectMcb(crossBrick, crossBrick) != Mcr.MCR_DISJOINT ) {
                // the volume of the intersection
                double crossVol = crossBrick.getVolume();
                // crossVol would only debit at 100% if the crown it grew into
                // was a rock--maple and birch aren't *that* big an impediment.
// 000802, here's the change
                crossVol *= 1.0 - shadower.getParams().lightCoeff;
                crossVol /= 2.0; // half for each

                // now find the volume of the column (inc. not intersecting)
                crossBrick.uz = z;
                crossBrick.lz = crownBottom;
                double overVol = crossBrick.getVolume();
                if (overVol == 0)
                  break;
                double debit = (overVol - crossVol) / overVol;
                lightbox.factorIn(crossBrick, debit );
                cropWannago.addDirection(x-shadower.x, y-shadower.y, debit);
              }
//            }
//          }
        }

        // second issue--trunk aversion
        // any babies (myage - theirage > 5 && theirage < 2) within
        // stem of me, I kill outright. Shadow is being reused here
        // as shadower's stem. Even if no chance of death, allow trunks to shove.
        double otherStemR = shadower.stemDiameter/(2.0 * 100.0);
        shadow2.circleToSquare(shadower.x, shadower.y, otherStemR);
        if ( shadow2.crossMcr(stem) ) {
          if ((shadower.getAge() < params.recruitAge)) {
            if (getAge() > params.recruitAge + 4) {
              shadower.die();
            } else { // both babies--I can move, but not kill
              cropWannago.addDirection(x - shadower.x, y - shadower.y, equiv.getArea());
            }
          // both big, both move.
          } else {
            cropWannago.addDirection(x - shadower.x, y - shadower.y, equiv.getArea());
          }
        }
      }
      somequad = somequad.pickWin(equiv, plantRoot, false);
    }
  }

  /** Override, 000803, appears mortMax and mortDecay model can be retired.
   *  But radius > stemRadius is too light a test, require 2x.
   */
  public boolean shallIDie() {
    // species-specific random death? apply before full-sun escape clause
    // use Woods global to turn on/off so don't need to screw with all the
    // Species definitions to turn off.
    if ( Woods.isSpeciesRandomDeath() ) {
      if ( Globals.getRandomSource().nextDouble() < params.mortRandom ) {
        return true;
      }
    }

    // if my crown is less than double r of trunk, die, because that's
    // getting ridiculous--careful, cm vs. m
    if ( ( radius * 100 ) <= stemDiameter)
      return true;

    // if pushed offsite, have no nutrients left, could have height 0....
    if ( z == 0)
      return true;

    // no more growth-dependent mortality--it's in the crown crop test
    return false;
  }

  /** 000803: override so that if no nutrients at trunk, factor = 0.
   */
  protected void findNutrientFactor() {
    // this lightlevel is per unit area of equiv area circle. Use same to
    // pick on the landscape.
    double nutrientArea = 0; // none until find some....
    McrCircle trunkMcr = new McrCircle();
    trunkMcr.circleToSquare(x, y, stemDiameter/(2.0*100));
    double trunkQuality = 0;

    Quad landRoot = Globals.getLandRoot();
    Quad somequad = landRoot.pickWin(equiv, landRoot, false);
    LimitResourceQuad limiter;
    while (somequad != null) {
      if (somequad instanceof LimitResourceQuad) {
        limiter = (LimitResourceQuad) somequad;
        if (equiv.intersectMcr(limiter, crossm) != Mcr.MCR_DISJOINT) {
          nutrientArea += limiter.getQuality(x, y) * crossm.getArea();
        }
        // require more than half of trunk on the site
        // without this, get ba-HEFTY edge effects, since edge trees not
        // subject to cropping, and nutFactor not yet having any effect.
        if (trunkMcr.intersectMcr(limiter, crossm) != Mcr.MCR_DISJOINT) {
          if (crossm.getArea() * 2.0 > trunkMcr.getArea())
            trunkQuality += limiter.getQuality(x, y);
        }
      }
      somequad = somequad.pickWin(equiv, landRoot, false);
    }

    // disallow any nutrients if trunk itself doesn't half cross the site.
    if (trunkQuality == 0) {
      nutrientFactor = 0;
      return;
    }

    nutrientFactor = nutrientArea/equiv.getArea();
    if (Double.isNaN(nutrientFactor))
      nutrientFactor = 0;
  }

}
