//Title:        Gecko Woods
//Version:
//Copyright:    Copyright (c) 2000, All rights reserved
//Author:       Ginger Booth, ginger@gingerbooth.com, gingerbooth@cyber-wizard.com
//Company:      Yale IBS/CCE
//Description:  "mini-Sortie"
/**
 * 000627, original: pull invariant species parameters out of Tree, to cut
 * memory needs. All fields are communicated via Tree, not here. Package-scope
 * publicity. Delegate record of Tree.
 */
package gecko.woods;
import gecko.*;
import java.awt.Color;

public class TreeParams {
  // defining parameters for the species
  // see p. 11, Pacala et al 1996 Ecological Monographs.
  double maxHeight;         // H1--meters
  double heightInitSlope;   // H2--meters/cm
  double crownToStem;       // C1--m/cm
  double crownDepthToHeight;// C1--m/m
  double lightCoeff;        // exp(-Ei), as given in table p. 11
  double growRateMax;       // G1
  double growRateMin;       // G2
  double growRingMax;       // G3--diameter rather than radial? Use 1/2 anyway...
  double mortMax;           // M1
  double mortDecay;         // M2
  double mortRandom;        // no name given - applies to the SPECIES
                                    // Reaper does global mortality games
  double recruitMeanD;      // mean distance parent->recruit (instead of R1)
  double bigFertility;      // R2
  double recruitRadius;     // no name given
  int recruitAge = 5;       // age at which newborn starts stepping
  // Sortie has no curb on crown growth--and applying their formulas, even for
  // finding maxCrownRadius at 95% height and fixing it, yields nonsense for
  // hemlocks. So, we need to add a parameter....
  double maxCrownRadius;    // curb on crown growth

  // derived from color (in Tree's superclass), but invariant for species.
  Color shadowColor;  // derived from primary color
  Color xorColor;     // derived from Driver background
}
