//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator
/**
 * 990929, vfb, original Java
 * Split this behavioral clump out of Critter for mix&match use for BacSim.
 */


package gecko.zoo;
import gecko.Globals;
import courseware.util.*;
import courseware.plot.*;
import courseware.jfcdep.*;

public class Direction {
  public double wannax, wannay, wannaz; // vector of direction to go--scratch space

  /** returns new double[3] as copy of wannax, wannay, wannaz after convert wannas
   *  to unit vector pointing same direction  */
  public double[] getDirection() {
    double[] d = Globals.ptToward(0, 0, 0, wannax, wannay, wannaz, 1.0);
    wannax = d[0];
    wannay = d[1];
    wannaz = d[2];
    return d;
  }
  public double getMagnitude2d() {
    return Math.sqrt((wannax * wannax) + (wannay * wannay));
  }
  public double getMagnitude3d() {
    return Math.sqrt((wannax * wannax) + (wannay * wannay) + (wannaz * wannaz));
  }
  /** return wannas as double[3], no normalizing first. */
  public double[] getWannas() {
    return new double[] { wannax, wannay, wannaz };
  }

  /** convert 2d wannas to unit vector pointing same direction, return angle of that.*/
  public double getDirectionAngle() {
    double wnorm;
    double angle = 0;

    wnorm = Math.sqrt(wannax * wannax + wannay * wannay); // hypotenuse...
    if (wnorm > 0.0) {
      wnorm = Math.abs(wannax / wnorm);
      if (wnorm > 1.0)
        wnorm = 1.0;
      angle = Math.acos(wnorm);// first quadrant angle...
      if (wannax < 0.0)
        angle = Math.PI - angle;
      if (wannay < 0.0)
        angle = (2.0 * Math.PI) - angle;
      wannax = Math.cos(angle);
      wannay = Math.sin(angle);
    }
    return angle;
  }

  public void setDirection(double dx, double dy) {
    wannax=dx;wannay=dy;
  }
  /* 2d */
  public void setDirection(double angle) {
    wannax=Math.cos(angle);
    wannay=Math.sin(angle);
  }
  public void zeroDirection() {
    wannax=0;wannay=0;wannaz=0;
  }
  /** convert dx, dy to unit vector before weighing in */
  public void addDirection(double dx, double dy, double weight) {
    double u[] = Globals.ptToward(0, 0, dx, dy, 1.0);
    wannax += u[0] * weight;
    wannay += u[1] * weight;
  }

  /** convert dx, dy, dz to unit vector before weighing in */
  public void addDirection(double dx, double dy, double dz, double weight) {
    double u[] = Globals.ptToward(0, 0, 0, dx, dy, dz, 1.0);
    wannax += u[0] * weight;
    wannay += u[1] * weight;
    wannaz += u[2] * weight;
  }

  public String toString() {
    return TruncVal.truncVal(wannax) + ", " + TruncVal.truncVal(wannay);
  }
}


