//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator
/**
 * 990617, vfb, original Java
 * Translated from the Swarm Objective-C Gecko source, version 0.51
 * With structural changes.
 */
/** Basic herbivore.
 */
package gecko.zoo;
import gecko.*;

public class Grazer extends SizedPlant {
  protected double lastmealx, lastmealy;	// where we last ate
  private double growrate;			// target growth rate
  private double veer;		    	// amount direction veers by
  protected double dirAngle;     // because direction gets cleared each time used, but my direction doesn't.
  private int    moverule;			// see getMoveRadius for interpretation
  protected static McrCircle huntwin = new McrCircle(); // avoid allocating the silly little things
  protected static McrCircle feedwin = new McrCircle();

  private static String[] myprops = { "growRate", "veer" };
  protected static String[] props;
  static {
    String[] firsthalf = SizedPlant.props;
    props = new String[firsthalf.length + myprops.length];
    for (int i=0; i<myprops.length; i++) {
      props[i] = myprops[i];
    }
    for (int i=0; i<firsthalf.length; i++) {
      props[myprops.length+i] = firsthalf[i];
    }
  }
  public String[] getPropertyNames() { return props;}


  public Grazer() {
    super();
    setPutOnPlantRoot(false); // go on mobileRoot
    dirAngle = 2 * Math.PI * Globals.getRandomSource().nextDouble();
    direction.setDirection(dirAngle);
    growrate = 0.0;
    getMetabolism().setIeat(GrassResources.RESOURCE_CARBOS);
    getMetabolism().setIam(GrassResources.RESOURCE_PROTEINS);
  }

  public void setGrowRate(double d) { growrate=d;}
  public void setVeer(double d) { veer=d;}
  public void setMoveRule(int i) { moverule=i;}
  public double getGrowRate() { return growrate;}
  public double getVeer() { return veer;}
  public int getMoveRule() { return moverule;}

  public void setFrom(Grazer other) {
    super.setFrom(other);
    growrate = other.growrate;
    veer = other.veer;
    moverule = other.moverule;
    dirAngle = 2 * Math.PI * Globals.getRandomSource().nextDouble();
    direction.setDirection(dirAngle);
  }

  public Speciesable makeKid() {
    Grazer kid = new Grazer();
    kid.setFrom(this);
    getSpecies().notifyBirthToMom(this);
// System.err.println("in makeKid, growrate=" + kid.growrate + ", direction/PI="
// + (kid.direction/Math.PI)+ ",newkid=" + Globals.agentID(kid));
    return kid;
  }

  public Object clone() {
    return makeKid();
  }

  public Move getDesiredMove() {
// System.err.println("... move direction/PI=" + (direction/Math.PI)
// + "," + Globals.agentID(this));
    Move move = new Move(dirAngle);
    move.multByDistance(getMoveRadius());
    return move;
  }

  protected double getMoveRadius() {
    double randomradius = getRadius() * Math.exp( Globals.getRandomSource().nextDouble() );
    switch (moverule) {
      case 2 : return        randomradius;
      case 3 : return 0.75 * randomradius;
      case 4 : return 0.75 * getRadius();
      case 5 : return 0.50 * randomradius;
      default :  // includes 1....
          return getRadius();
    }
  }


  /** Choose a plant--kinda arbitrary logic, so localize.

   * Don't care how big plant is, so long as it's bigger than minradius
   * but will return plant smaller than minradius, if can't find bigger.
   */
  protected Critter bestpreyMin(double minradius) {

    Quad pq;
    Critter plant, bestplant, anyplant;
    Quad plantRoot = Globals.getPlantRoot();

    feedwin.circleToSquare(getX(), getY(), getRadius());
    pq = plantRoot.pickWin(feedwin, plantRoot, false);
    bestplant = null;	// any plant that matches minradius
    anyplant  = null;	// some plant that doesn't match minradius
    
    if (pq != null) do {
        plant = (Critter) pq;
        if (this.eats(plant.getMetabolism().getIam())) {
            if (plant.getRadius() > minradius) {
                if (Globals.getRandomSource().nextDouble() > 0.5)
                    bestplant = plant;
                else if (bestplant == null)
                    bestplant = plant;
            } else if (bestplant==null) {
                if (anyplant==null)
                    anyplant = plant;
                else if (Globals.getRandomSource().nextDouble() > 0.5)
                    anyplant = plant;
            }
        }
        pq = pq.pickWin(feedwin, plantRoot, false);
    } while (pq != null);

    if (bestplant != null)
        return bestplant;
    else
        return anyplant;
  }

  /** we have a memory of direction in addition to super's current wanna
   *  direction, which is wiped out each step. */
  public void setDirection(double angle) {
    direction.setDirection(angle);
    dirAngle = angle;
  }
  /** we have a memory of direction in addition to super's current wanna
   *  direction, which is wiped out each step. */
  public void setDirection(double tox, double toy) {
    direction.setDirection(tox, toy);
    dirAngle = direction.getDirectionAngle();
  }

  /** How much I want to eat depends on -my- size & tax debt & growthrate.
   *  Hunt for any plant that satisfies that desire, settle for smaller
   *  one if can't have what we want.
   *
   *  vfb, 990907, set z based on what grazer just ate. Only matters to POVCamera
   *  for now.
   */
  public void hunt() {
    Critter plant;
    int me = getMetabolism().getIam();
    int meat = getMetabolism().getIeat();
    double mybite;
    double mybiter;
    double angle, veerangle;
    double myvol;

    myvol   = getMetabolism().getStores(me);
    mybite  = ( getMetabolism().getTaxRate(me) * Math.pow( myvol, getMetabolism().getTaxExponent(me) ) )
              + growrate * myvol;
    mybite /= getMetabolism().getConversion( getMetabolism().getIeat(), me );
    mybiter = Globals.volume2radius(mybite, 1.0, getScaleVolume());

    plant = bestpreyMin(mybiter);

    if (plant!=null) {
        if (plant.getRadius() > mybiter) {
            plant.stealResource( meat, mybite );
        } else {
            mybite = plant.stealAllResource(meat);
        }
        plant.resetRadius();  // since another may choose to eat him.
        this.feed(meat, mybite);
        setZ((plant.getRadius() - getRadius()) * 1.1);

        // move at semi-random, save last place we ate
        // angle = (M_PI / 8.0) * [randSource evenRand];
        // if ([randSource evenRand] > 0.5) angle = -angle;
        angle = dirAngle;
        veerangle = (veer/2.0) * Globals.getRandomSource().nextDouble();
        if (Globals.getRandomSource().nextDouble() < 0.5)
            veerangle = - veerangle;
        angle += veerangle;
        setDirection( angle );
        lastmealx = plant.getX();
        lastmealy = plant.getY();

    } else {
        setZ(getRadius() * 1.1);

        // back up toward last place we found food
        this.setDirection( lastmealx - this.getX(), lastmealy - this.getY());
    }
  }


  public String toString() {
    return "Grazer[veer="+veer+",growrate="+growrate+",veer="+veer+

           ",super=" + super.toString() + "]";

  }


}


