//Title:        Gecko
//Version:
//Copyright:    Copyright (c) 1999, All rights reserved
//Author:       Ginger Booth, gingerbooth@cyber-wizard.com, ginger@gbooth.com
//Company:      Yale IBS/CCE
//Description:  Individual-Based Ecology Simulator
/**
 * 000201, vfb, rewrite for new ScaredySituation changes to ScaredyGrazer.
 *
 * 000117, vfb, Ofer's changes for state-based decision making on recent eating
 * success, progress toward life reprod. goals, & recent predator sightings.
 * Also fix so stay extinct in multi-season runs but restore to original initNum
 * on restart. Few more stat output files--track deaths and cause of death, and
 * track individual progress on sample of individuals.
 *
 * 990930, vfb, original Java
 * Ofer's variant on Os' ScaredyGrazers. First changes: reproduce in bulk at
 * end of season.
 *
 * 001214, vfb, changes per Ofer's sit-and-wait spider.
 */


package gecko.zoo;
import gecko.*;
import gecko.statglue.*;
import courseware.jfcdep.TruncVal;
import java.io.*;
import stats.*;

public class OferStateGrazer extends ScaredyGrazer {
  private double kidCostFactor; // debit to adult is this * kidvolume
  // HSC = half-saturation constant, shapes probability function for
  // this factor in decisions
  private double kidGoal;       // how many kids used for progress toward goal
  private double kidGoalExponent; // meeting life size/reprod goal trajectory
  // private int predatorMemLength;// memory length for predator encounters
  // private double predatorHSC;   // recent predator encounters
  // this one uses %satisfaction^exponent for better shaped probabilities
  private double hungerExponent;     // meeting eating goals last round or two
  private double breedvol, afterbreedvol, kidcost; // used in progress and reproduce

  // for the first numTracers born of this species, traceLog is non-null
  private int numTracers;
  private PrintWriter traceLog = null;

  // one deathLog per Run (not Season, but Run)
  static protected PrintWriter deathLog;

  private static String[] myprops = { "kidCostFactor",
    "kidGoal", "kidGoalExponent",
    // "predatorMemLength", "predatorHSC",
    "hungerExponent", "numTracers"};
  protected static String[] props;
  static {
    String[] firsthalf = ScaredyGrazer.props;
    // tricky bit==omit "adultTime"
    props = new String[firsthalf.length + myprops.length - 1];
    for (int i=0; i<myprops.length; i++) {
      props[i] = myprops[i];
    }
    int skipper = 0;
    for (int i=0; i<firsthalf.length; i++) {
      if (!firsthalf[i].equals("adultTime")) {
        props[myprops.length+skipper] = firsthalf[i];
        skipper++;
      }
    }
  }
  public String[] getPropertyNames() { return props;}

  // working vars
  // private boolean[] predatorMemory;
  private double[] threatXYD = new double[3]; // need often...
  private double lastMealSatisfaction;
  private double prevMealSatisfaction;

  public void setKidCostFactor(double kcf) {kidCostFactor = kcf;}
  public double getKidCostFactor() {return kidCostFactor;}
  public void setBreedSize(double d) {breedsize = d;}
  public double getBreedSize() {return breedsize;}
  // public int getPredatorMemLength(){return predatorMemLength;}
  // public void setPredatorMemLength(int v) {predatorMemLength = v;}

  public double getKidGoal(){return kidGoal;}
  public void setKidGoal(double v) {kidGoal = v;}
  public double getKidGoalExponent(){return kidGoalExponent;}
  public void setKidGoalExponent(double v) {kidGoalExponent = v;}
  // public double getPredatorHSC(){return predatorHSC;}
  // public void setPredatorHSC(double v) {predatorHSC = v;}
  public double getHungerExponent(){return hungerExponent;}
  public void setHungerExponent(double v) {hungerExponent = v;}

  /** Because there are no new parameters to OferStateGrazer, we don't need
   *  most of the creation rigmarole, but we do need makeKid, or it won't
   *  turn out of the right class, so won't call the right (new) behavior. */
  public OferStateGrazer() {super();}

  /** set memory of predator sighted yes/no this timestep. */
  /*
  protected void rememberSawPredator() {
    if (predatorMemLength==0) return;
    int index = (int) Globals.getCurrentTime() % predatorMemLength;
    predatorMemory[index] = ( searchForThreat(getSmellRadius(), threatXYD) );
  }
  protected boolean seePredator() {
    int index = (int) Globals.getCurrentTime() % predatorMemLength;
    return predatorMemory[index];
  }
  */
  /** Percent of predatorMemLength recent timesteps where predator visible.*/
  /*
  protected double predatorPercentLately() {
    double count = 0;
    for (int i=0; i<predatorMemLength; i++)
      if (predatorMemory[i])
        count++;
    return count/((double)predatorMemLength);
  }
  */

  public void setFrom(OferStateGrazer other) {
    super.setFrom(other);
    breedsize = other.breedsize;
    kidCostFactor = other.kidCostFactor;
    kidGoal = other.kidGoal;
    kidGoalExponent = other.kidGoalExponent;
    // predatorMemLength = other.predatorMemLength;
    // predatorHSC = other.predatorHSC;
    hungerExponent = other.hungerExponent;
    numTracers = other.numTracers;
    // initialize stuff for decision making
    // predatorMemory = new boolean[predatorMemLength]; // false'd by defn
    breedvol = Globals.radius2volume(breedsize, getScaleRadius(), getScaleVolume());
    ResourceList cost = getKidCost(getKidShare());
    kidcost = cost.list[getMetabolism().getIam()];
    afterbreedvol = breedvol - kidcost;
  }

 private double targetKidsProgress() {
    int timeleft = (int) (getMaxAge() - 1 - getAge());
    // ok, that was total time left. But let's subtract off the time needed to
    // get from here to afterbreedvol the first time (or maybe another time...)
    double vol = Globals.radius2volume(radius, getScaleRadius(), getScaleVolume());
    double equivContinuousRate = Math.log(1.0 + getGrowRate());
    if (vol < afterbreedvol) {
      double timeToSize = Math.log(afterbreedvol/vol)/equivContinuousRate;
      timeleft -= timeToSize;
    }
    // OK, now the time in timeLeft is used solely for getting from
    // afterbreedvol to breedsize, stopping for one timestep to breed, and
    // doing it again. For however many babies we project. So, this
    // timeToSize is the time for one breeding event.
    double timeToSize = 1 + Math.log(breedvol/afterbreedvol)/equivContinuousRate;
    double numbabies = kidcount + timeleft/timeToSize;
    double progress = numbabies/kidGoal;
    if (progress > 1.0)
      progress = 1.0;
    else if (progress < 0)
      progress = 0.0;

  /** Project reproductive goal percent realization. Match reproduce().
   */
     // To solve for t, find the equivalent continuous r = ln(L).
    // Then can substitute exp(r) for L, take the log of both sides of equation,
    // and solve for t.
    /*
    double equivContinuousRate = Math.log(1.0 + getGrowRate());
    double timeToSize = Math.log(targetVolume/vol)/equivContinuousRate;
    double timeprogress = ((double) timeleft) - timeToSize;
    if (timeprogress < 0)
      System.err.println("time " + Globals.getCurrentTime() +
      ", timeprogress=" + TruncVal.truncVal(timeprogress));
    */

    return progress;
  }


  public Speciesable makeKid() {
    OferStateGrazer kid = new OferStateGrazer();
    kid.setFrom(this);
    kid.initKidMood();
    getSpecies().notifyBirthToMom(this);

    // once per RUN (not season) set up the deathLog
    // is in the way less of the time here than in step.
    // getSpecies().isNewRun() returns true once per run.
    if (getSpecies().isNewRun()) {
      createDeathLog();
    }

    // if the population is less than numTracers, I get
    // a tracelog.
    if (getSpecies().getPopulation() < numTracers)
      kid.createTraceLog();

    return kid;
  }
  public Object clone() {return makeKid(); }


  /** Critter reproduce resets radius--want all babies to say they

   *  were born to Mom of starting radius, so want no reset radius

   *  until end, and can't go by radius to see if more kids to make. */

  public void oneBaby(ResourceList kidgets, ResourceList cost) {
    Critter baby;
    double distance;
    double angle;
    Metabolism metabolism = getMetabolism();

    if (kidcount == 0)
      firstbornon = Globals.getCurrentTime();
    kidcount++;

    angle = 2.0 * Math.PI * Globals.getRandomSource().nextDouble();
    // make the kid
    Speciesable lbaby = makeKid();
    baby = (Critter) lbaby;
    distance    = getKidDistance();
    baby.x      = x + ( distance * Math.cos(angle) );
    baby.y      = y + ( distance * Math.sin(angle) );
    baby.direction.setDirection(angle);

    // transfer resources
    metabolism.debit(cost);
    baby.getMetabolism().setStores(kidgets);
    baby.resetRadius();

    // hang on quad tree
    baby.setMcr();
    Quad myroot = getRoot();
    myroot.addNewbie(baby);
    getSpecies().notifyBirth();
  }


  public boolean wantsToReproduce() {

    return (radius >= breedsize);

  }

  protected ResourceList getKidCost(ResourceList kidgets) {
    ResourceList cost = new ResourceList();
    cost.setFrom(kidgets);
    cost.multiply(kidCostFactor);
    return cost;
  }

  /** Superclass wantsToReproduce and reproduce do a lot of
   *  important things, like set breedsize of time limited critter.
   *  Note this routine must match findTargetKidsProgress.
   */
  public boolean reproduce() {
    boolean did = false;

    if (wantsToReproduce()) {
      int numbabies;
      double scaleradius = getScaleRadius();
      double scalevolume = getScaleVolume();
      double vol = Globals.radius2volume(radius, scaleradius, scalevolume);
      numbabies = Math.round(Math.round(Math.floor((vol - afterbreedvol) / kidcost)));
      ResourceList kidgets = getKidShare();
      ResourceList cost = getKidCost(kidgets);

      /* v. useful debug.
      System.err.println(getSpecies().getName() + " babies=" + numbabies +
      ", time progress=" + TruncVal.truncVal(targetKidsProgress()) +
      ", I risked it " + merisk + " / " + meAllRisk +
      ", overall=" + didRiskCount + " / " + riskPossible);
      */
      for (int i=0; i<numbabies; i++) {
        oneBaby(kidgets, cost);
        did = true;
      }
      resetRadius();
    }
    if (did)
      lastMealSatisfaction = 0; // don't eat and reproduce on same round
    return did;
  }

  /** remember % ideal meal I last managed to get. Overrides super. */
  protected double eat() {
    double gotpercent = super.eat();
    lastMealSatisfaction = gotpercent;
    return gotpercent;
  }

  /** override move limits per Ofer request--can move by smellR if threatened. */
  private void limitMove() {
    if (wannadistance <= 0.0)
      wannadistance = 0;
    // no other limit
  }

  /** tricky bit of "logic" for new threat, if happy.
   *  The returned boolean is true if need to miss dinner.
   */
  private boolean happyMissDinner()
  {
    boolean haveGrass;
    if (situation.amHidden) {
      // if safe and have grass, stay put, mood stays happy
      // NOTE: ScaredySituation will NOT set yumPlant to preferred grass
      // if I'm threatened. But let's leave the logic here should that
      // get switched the other way in ScaredySituation. Basically, decided
      // that under field conditions, they really don't eat preferred grass
      // while hiding. That's a geometric artifact we shouldn't make use of.
      haveGrass = ( ( situation.yumPlant != null ) &&
                  ( situation.yumPlant.getSpecies() != getRefuge() ) );
      if (haveGrass) {
        // orient self away from spider, but go nowhere
        // though not running, the mood is run, so don't freeze later
        setIntendedMove(RUN_FROM_THREAT);
        wannadistance = 0;
      // have no grass, want to get risky?
      } else {
        // this isn't true--we're not running, but since we are leaving
        // our safety by choice, make sure we don't freeze next round.
        // otherwise, we've just made a really bad choice, because if we
        // freeze next round, we'll net lose bodymass.
        mood = SCAREDY_RUN;
        if (riskIt()) {
          // can I go directly toward food?
          haveGrass = ( (situation.farYumPlant != null) &&
                        (situation.farYumPlant.getSpecies() != getRefuge()) );
          if (haveGrass) {
            setIntendedMove(RUN_TO_FAR_BEST);
            // note that this is -after- limitMove, so overrides the limits
// wannadistance = situation.farYumD;
          // we'll just go away from spider and hope prospects improve
          } else {
            setIntendedMove(RUN_FROM_THREAT);
// wannadistance = radius * getSmellRadius();
          }
        } else {
          runOrHide();
        }
      }
      return false;   // go ahead and eat

    } else { // threatened and not safe and wish to be
      mood = SCAREDY_FROZE;
      // orient away from threat, but don't move.
      setIntendedMove(RUN_FROM_THREAT);
      wannadistance = 0;
      lastMealSatisfaction = prevMealSatisfaction;
      return true;                                       // MISS DINNER
    }
  }

  /** return true if get no dinner */
  private boolean frozenMissDinner() {
    // our mood is not ever FROZE if were safe already....
    mood = SCAREDY_RUN; // mood progresses in any case
    if (riskIt()) {
      // just move away from the spider and carry on normally
      setIntendedMove(RUN_FROM_THREAT);           // eat
// wannadistance = radius * getSmellRadius();
      return false;
    } else {
      // miss dinner, run for safety
      // no check for closer predator as per ScaredyGrazer case
      runOrHide();
      return true;                                // MISS DINNER
    }
  }

  public void runRisky() {
    boolean haveGrass;
    if (riskIt()) {
      // can I go directly toward food?
      haveGrass = ( (situation.farYumPlant != null) &&
                    (situation.farYumPlant.getSpecies() != getRefuge()) );
      if (haveGrass) {
        setIntendedMove(RUN_TO_FAR_BEST);
        // note that this is -after- limitMove, so overrides the limits
// wannadistance = situation.farYumD;
      // we'll just go away from spider and hope prospects improve
      } else {
        setIntendedMove(RUN_FROM_THREAT);
// wannadistance = radius * getSmellRadius();
      }
    } else {
      runOrHide();
    }
  }

  public void hunt() {
    // eggs don't hunt
    if (mood == SCAREDY_EGG)
      return;

    // ---------------------------------------------------------------
    // 1. set mood and intended next move based on current perceptions
    // ---------------------------------------------------------------
    // this looks for all my options and threats
    situation.lookAround(this, isFearless());
    wannadistance = this.getRadius(); // unless proven otherwise
    prevMealSatisfaction = lastMealSatisfaction;

    if (situation.threat == null) {
      mood = SCAREDY_HAPPY;
      setIntendedMoveRandom();

    // if I am threatened, my actions (including whether I get to eat)
    // depend on a. my mood (experience last step) and b. refuge availability
    } else {
      // rem eggs exited at first line of routine
      switch (mood) {

        case SCAREDY_HAPPY :
          if (happyMissDinner())
            return;                                     // maybe MISS DINNER
          break;

        case SCAREDY_FROZE :
          if (frozenMissDinner())
            return;                                     // maybe miss dinner
          break;

        case SCAREDY_RUN :
          runRisky();
          // we don't miss dinner again
          break;
      }
    }

    // -------------------------------------------------------------------
    // 2. figure out what/how much to eat, if I didn't already miss dinner
    // -------------------------------------------------------------------
    eat();
  }



  /** Apply our probabilities from our 3 considerations: kidgoal progress,
   *  recent meal dissatisfaction, and recent predator sightings
   */
static int riskItCalled = 0;
static int riskPossible = 0;
static int didRiskCount = 0;
int merisk = 0;
int meAllRisk = 0;
  protected boolean riskIt() {
    // these are all expressed as probability of seeking safety, so invert
    // the probability at the end
riskItCalled++;
meAllRisk++;
    double psafe = 0;
    double psat  = 0;
    double pkid  = 0;

    // how many probabilities are in effect? non-zero exponents
    double numberOfFactors = 0;

    // satisfaction 1.0->probability of seeking safety = 1.0
    // satisfaction uses different shaped curve due to noncontinuous values
    if (hungerExponent != 0) {
      numberOfFactors += 1.0;
      psat = Math.pow(lastMealSatisfaction, hungerExponent);
      psafe += psat != 0 ? (1.0/psat) : 0;
    }
    if (kidGoalExponent != 0) {
      numberOfFactors += 1.0;
      pkid = Math.pow(targetKidsProgress(), kidGoalExponent);
      psafe += pkid != 0 ? (1.0/pkid) : 0;
    }
    // predators every step = 1.0 -> probability of seeking safety = 1.0
    // double ppred = probabilityCurve(predatorPercentLately(), predatorHSC);

    // solve for harmonic mean:
    psafe  = 1.0/psafe;
    psafe *= numberOfFactors;

    // invert probability of seeking safety to probability of risking forth
    double prisk = 1.0 - psafe;
if (prisk > 0) 
riskPossible++;

    boolean doIt = Globals.getRandomDouble() < prisk;
if (doIt) {
didRiskCount++;
merisk++;
}
    return doIt;
  }

  /*
  public double probabilityCurve(double praw, double halfSatConstant) {
    return praw / (halfSatConstant + praw);
  }
  */

  protected void createDeathLog() {
    BatchUI batchUI = (BatchUI) Globals.getUI();
    String filename = batchUI.getFileStem() + "_deathlog.txt";
    try {
      deathLog = batchUI.openFile(filename);
    } catch (Exception e) {}
    if (deathLog != null)
      deathLog.println("# time species 1_killed_or_0_starved");
  }

  protected void createTraceLog() {
    BatchUI batchUI = (BatchUI) Globals.getUI();
    int id = Math.round(Math.round(getSpecies().getPopulation()));
    String filename = batchUI.getFileStem() + "_" + getSpecies().getName()
      + id + ".txt";
    try {
      traceLog = batchUI.openFile(filename);
    } catch (Exception e) {}
    if (traceLog == null) return;
    traceLog.println("# species = " + getSpecies().getName() +
      ", member " + id);
    traceLog.println("# time lastMealSatisfaction spidersNow kidGoalProgess numKids risksTaken radius");
  }

  protected void writeTrace() {
    traceLog.println(
      Globals.getCurrentTime() + " " +
      TruncVal.truncVal(lastMealSatisfaction) + " " +
      situation.threatCount + " " +
      TruncVal.truncVal(targetKidsProgress()) + " " +
      kidcount + " " +
      merisk + " " +
      TruncVal.truncVal(radius)
    );
  }

  protected void writeEpitaph() {
    if (deathLog == null) return;
    // time species 1_killed_or_0_starved
    int deathtype = killedFlag ? 1 : 0;
    deathLog.println(Globals.getCurrentTime() + " " +
      getSpecies().getName() + " " + deathtype);
    if (traceLog != null) {
      traceLog.print("# died of ");
      if (killedFlag) traceLog.println(" predation");
      else  traceLog.println(" starvation");
      traceLog = null;
    }
  }

  private boolean killedFlag = false;
  /** Carnivores call this one to kill. Override for output flag. */
  public double stealAllResource(int resourceType) {
    killedFlag = true;
    return super.stealAllResource(resourceType);
  }

  /** override for deathLog output. False=Critter died.*/
  public boolean paytax() {
    boolean verdict = super.paytax();
    if (!verdict)
      writeEpitaph();
    return verdict;
  }

  /** This is the number of tracers to be used on this Species (like, GopperMedium).*/
  public void setNumTracers(int n) {numTracers = n;}
  public int getNumTracers() {return numTracers;}

  /** Need to override this one because want stages tallied before reproduce... */
  public void step() {
   long today = Globals.getCurrentTime();
   // if one timestep past birthday for season, resetInitNum
   if ( getSpecies().isBirthday() ) {
      getSpecies().setWorkingInitNumber(0);

riskItCalled = 0;
riskPossible = 0;
didRiskCount = 0;
merisk = 0;
meAllRisk = 0;

   }


   // always look and remember current predator-around experience

   // rememberSawPredator();


   if ( (today > 0) && (getAge() == 0) )

     return;

    // ScaredyGrazer's contribution to step
    getSpecies().notifyAlive(this); // for stages

    if (mood == SCAREDY_EGG) {
      if (seasonCheck()) ; // just checking it does what must be done.
      getSpecies().notifyBiomass(getMetabolism().getStores(getMetabolism().getIam()));
      return;
    }
    if (deathdayCheck())
      return;
    if (seasonCheck())
      return;
    metabolize();
    if (!paytax()) {
      return;
    }
    if (!reproduce()) {
      hunt();
    }
    move();

    // stat reports
    if (traceLog != null)
      writeTrace();
    getSpecies().notifyBiomass(getMetabolism().getStores(getMetabolism().getIam()));
  }
}



