
//Title:        Indeco Steel Simulator
//Version:      
//Copyright:    Copyright (c) 1999
//Author:       Joseph Guse & Ginger Booth
//Company:      Yale FES/CCE
//Description:  Individual-based economic simulator
//of the steel industry, ore->consumer,
//with transportation.

/** ApplicationGUI does both graphics and log files. */

/** Adopted from Indeco for Gecko-in-CourseWare, 990708, vfb. */

package stats;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;

import courseware.util.ParamList;

public class ApplicationGUI extends BatchUI {
  private Vector vizPlotFrames; // currently visible PlotFrames
  private Vector miscFrames;    // arbitrary frames
  private static int tilingOffset; // for positioning next frame
  private static int framecount;   // for positioning next frame
  private Frame masterFrame;
  private Timer timer;
  private boolean firstLayout;

  public ApplicationGUI(String fileStem, String protocolFile,
                        Frame masterFrame, Timer timer) {
    super(fileStem, protocolFile, timer);
    this.masterFrame = masterFrame;
    vizPlotFrames = new Vector(1,1);
    miscFrames = new Vector(1,1);
    firstLayout = true;
  }

  public void bindParams(ParamList pl) {
    pl.bindAllObjects(this.getClass(), this);
  }

  public Frame getFrame() {return masterFrame;}

  public boolean isGUI() { return true;}

  /** Hooks statistic "plot" to new output file. */
  synchronized public void makeStatDumpers(MetaPlot metaplot) {
    MetaSink ms = getMetaSink(metaplot);
    if ((ms != null) && (ms.hasStatDumper(PlotFrame.class)))
      return;

    if (ms == null)
      super.makeStatDumpers(metaplot);

    metaplot.setMayTruncateTraces(false); // batch super set to true
    PlotFrame pf = new PlotFrame(metaplot.getName());
    if (metaplot.getNeedLegend())
      pf.plotter.setPlotLegend(true);
    displayFrame(pf);
    metaplot.showOn(pf);
    ms = getMetaSink(metaplot);
    if (ms == null) {
      ms = new MetaSink(metaplot);
    }

    ms.addStatDumper(pf);

    class WinCloser extends WindowAdapter {
      Frame theFrame;
      MetaSink theMetaSink;
      public WinCloser(Frame pf, MetaSink ms) {theFrame=pf; theMetaSink=ms;}
      public void windowClosing(WindowEvent e){
        theMetaSink.removeStatDumper((StatDumper)theFrame);
      }
    }
    pf.addWindowListener(new WinCloser(pf, ms));
  }

  public void displayFrame(Frame frame) {
    miscFrames.addElement(frame);
    showFrame(frame);
    class WinCloser extends WindowAdapter {
      Frame theFrame;
      public WinCloser(Frame frame) {theFrame=frame;}
      public void windowClosing(WindowEvent e){
        miscFrames.removeElement(theFrame);
      }
    }
    frame.addWindowListener(new WinCloser(frame));
  }

  public void displayFrame(PlotFrame frame) {
    if (!vizPlotFrames.contains(frame))
      vizPlotFrames.addElement(frame);
    showFrame(frame);
    class WinCloser extends WindowAdapter {
      PlotFrame theFrame;
      public WinCloser(PlotFrame frame) {theFrame=frame;}
      public void windowClosing(WindowEvent e){
        vizPlotFrames.removeElement(theFrame);
        for (int i=0; i<getMetaSinks().size(); i++) {
          MetaSink ms = (MetaSink)getMetaSinks().elementAt(i);
          if (ms.hasStatDumper(theFrame.plotter)) {
            ms.removeStatDumper(theFrame.plotter);
          }
        }
      }
    }
    frame.addWindowListener(new WinCloser(frame));
  }

  private void showFrame(Frame frame) {
    Point loc;
    if ( (firstLayout) && (masterFrame != null) ) {
      loc = masterFrame.getLocationOnScreen();
      tilingOffset = masterFrame.getSize().height;
    } else {
      loc = new Point(0,0);
      tilingOffset = 25;
    }
    firstLayout = false; // whatever we started doing, keep it up.
    int offset = framecount * tilingOffset;
    framecount++;
    loc.x += offset;
    loc.y += tilingOffset;
    frame.setLocation(loc);
    frame.pack();
//    frame.setVisible(true);
  }

  /** Do whatever's necessary to clean up Between Runs. */
  public void clear() {
    vizPlotFrames = new Vector(1,1);
    framecount = 0;
    super.clear();
  }
  /** Do whatever's necessary to restart this run. */
  public void reset() {
    for (int i=0; i<vizPlotFrames.size(); i++)
      ((PlotFrame) vizPlotFrames.elementAt(i)).reset();
    super.reset();
  }

  /** Do whatever's necessary to clean up at End. */
  public void finalize() {
    for (int i=0; i<vizPlotFrames.size(); i++) {
      if (vizPlotFrames.elementAt(i) instanceof PlotFrame)
        ((PlotFrame) vizPlotFrames.elementAt(i)).finalize();
      else
        System.err.println("AppGUI.finalize, not plotframe: " + vizPlotFrames.elementAt(i));
    }
    super.finalize();
  }

}
