
//Title:        Indeco Steel Simulator
//Version:      
//Copyright:    Copyright (c) 1999
//Author:       Joseph Guse & Ginger Booth
//Company:      Yale FES/CCE
//Description:  Individual-based economic simulator
//of the steel industry, ore->consumer,
//with transportation.

/** Trace weighted average of member statbrokers. If you don't want
 *  it weighted, tell your statbrokers to weight at 1.0 before
 *  adding the statbrokers to the trace.
 */

/** Adopted from Indeco for Gecko-in-CourseWare, 990708, vfb. */

package stats;


import java.awt.Color;

public class AvgTrace extends SumTrace {

  public AvgTrace(String title) { super(title); }
  public AvgTrace() { super(); }
  public AvgTrace(String name, Color color, Timer timer) { super(name, color, timer); }

  public double doSum() {
    if (members == null)
      return 0.0;
    double sum = 0;
    double wsum = 0;
    int nvalid = 0;
    StatBrokerable broker;
    double value, weight;
    for (int i=0; i<members.size(); i++) {
      broker = (StatBrokerable) members.elementAt(i);
      if ( !( invalid( value=broker.getValue() ) ) ) {
        if ( !( invalid( weight=broker.getWeight() ) ) ) {
          sum += value * weight * broker.getNormalizer();
          wsum += weight;
          nvalid++;
        }
      }
    }
    // need to distinguish between legitimate sum of 0.0 and
    // no-valid-values-found
    double answer;
    if (nvalid == 0)
      answer = StatBrokerable.INVALID_STAT;
    else if (wsum == 0)
      answer = 0;
    else
      answer = sum / wsum;
    return answer;
  }

  public void step() {
    super.add(doSum(membersX), doSum());
  }

  /** atIndex should basically be "index of now", zero at the beginning. */
  public void start() {
    super.setX(0, doSum(membersX));
    super.setX(0, doSum());
  }


  public String toString(){
    return "AvgTrace" + getLabel();
  }

}
