
//Title:        Indeco Steel Simulator
//Version:      
//Copyright:    Copyright (c) 1999
//Author:       Joseph Guse & Ginger Booth
//Company:      Yale FES/CCE
//Description:  Individual-based economic simulator
//of the steel industry, ore->consumer,
//with transportation.

/** Write stats to file. Since each line starts with the time now, this
 *  class needs a Timer. That interface defines "int getCurrentTime()".
 */

/** Adopted from Indeco for Gecko-in-CourseWare, 990708, vfb. */

package stats;

import java.io.*;
import java.util.*;
import courseware.plot.*;


public class FileStatDumper implements StatDumper {
  private PrintWriter output;
  private Vector traces;
  private boolean firstLineWritten;
  private final static char separator = ',';
  private final static String runSeparator = " "; // blank line
  private final static String commentPrefix = "";
  private Timer timer;

  public FileStatDumper(PrintWriter output, Timer timer) {
    this.output = output;
    firstLineWritten = false;
    traces = new Vector(5,5);
    this.timer = timer;
  }
  public String toString() {
    return "FileStatDumper[traces.size=" + traces.size() + ",output=" + output + "]";
  }
  public int addTrace(SumTrace readyTrace) {
    traces.addElement(readyTrace);
    return traces.size() - 1;
  }

  /** We DON'T want traces getting scrambled. Null placeholder inserted. */
  public void removeTrace(SumTrace readyTrace) {
    int wasTrace = traces.indexOf(readyTrace);
    if (wasTrace > -1)
      traces.setElementAt(null, wasTrace);
  }

  public void clear() {
    traces = new Vector(5,5);
    output.println(runSeparator);
  }

  private void labelColumns() {
    SumTrace trace;
    output.print("Time");
    output.print(separator);

    for (int traceId=0; traceId < traces.size() - 1; traceId++) {
      trace = (SumTrace) traces.elementAt(traceId);
      if (trace != null)
        if (trace.getLabel() != null)
          output.print(trace.getLabel());
      output.print(separator);
    }
    if (traces.size() > 1) {
      trace = (SumTrace) traces.elementAt(traces.size() - 1);
      if (trace != null) {
        if (trace.getLabel() != null)
          output.print(trace.getLabel());
      }
    }
    output.println();
  }

  private void dumpLine() {
    SumTrace trace;
    output.print(timer.getCurrentTime());
    output.print(separator);
    for (int traceId=0; traceId < traces.size(); traceId++) {
      trace = (SumTrace) traces.elementAt(traceId);
      if (trace != null) {
        if (!trace.isXTime()) {
          output.print(trace.getLastX());
          output.print(separator);
        }
        output.print(trace.getLastY());
      }
      if ( traceId < (traces.size() - 1) )
        output.print(separator);
    }
    output.println();
  }

  public void step() {
    if (traces.size() < 1)
      return;
    if (!firstLineWritten) {
      labelColumns();
      firstLineWritten = true;
    }
    dumpLine();
  }

  /** Doesn't do anything. A new column appears.... */
  public void update() {
  }

  public void finalize() {
    output.flush();
    output.close();
  }
}