
//Title:        CourseWare stats package
//Version:      
//Copyright:    Copyright (c) 1998
//Author:       Ginger Booth
//Company:      CourseWare/CCE/YIBS
//Description:  CourseWare/Gecko/Indeco stats machinery

/** We need a few behaviors added to Plot. Differs wildly from Indeco's
 *  Plotter, which was a Swing-modified-then-extended source copy of Plot.
 */


package stats;

import courseware.util.*;
import courseware.plot.*;
import courseware.jfcdep.*;
import java.awt.*;
import java.awt.event.*;

public class Plotter extends Plot implements StatDumper {
  private static int defsizex = 400;
  private static int defsizey = 240;
  private static Color defBackgroundColor = Color.white;
  private static Color defForegroundColor = Color.black;
  private static Color defGridColor = Color.lightGray;

  public Plotter() {this(defsizex, defsizey);}

	public Plotter(int graphWidth, int graphHeight) {
    super(new Dimension(graphWidth, graphHeight));
    setForeground(defForegroundColor);
    setBackground(defBackgroundColor);
    setGridColor(defGridColor);
	}

  /** All plotters are created at default size. */
  static public void setDefSizeX(int i){defsizex = i;}
  static public void setDefSizeY(int i){defsizey = i;}
  static public void setDefBackgroundColor(Color c) {defBackgroundColor = c;}
  static public void setDefForegroundColor(Color c) {defForegroundColor = c;}
  static public void setDefGridColor(Color c) {defGridColor = c;}

  public int addTrace(SumTrace readyTrace) {
    return super.addTrace(readyTrace);
  }
  public void removeTrace(SumTrace readyTrace) {
    int which = data.indexOf(readyTrace);
    if (which > -1) {
      data.removeElementAt(which);
    }
  }

  public void step() {
    double xmax, ymax, ymin, z;
    xmax = ymax = ymin = 0.0;
    SumTrace st;
    for (int i=0; i<data.size(); i++) {
      st = (SumTrace)data.elementAt(i);
      st.step();
      if ((z = st.getLastX()) > xmax) xmax = z;
      if ((z = st.getLastY()) > ymax) ymax = z;
      if ((z = st.getLastY()) < ymin) ymin = z;
    }
    if (getXmax() < xmax) setXmax(xmax);
    if (getYmax() < ymax) setYmax(ymax);
    if (getYmin() > ymin) setYmin(ymin);
    setNeedRepaint(true);
  }
  public void reset() {
    setXmin(0); setYmin(0);
    setXmax(.1); setYmax(.1);
    update();
  }

  /** Print an object as a unique id for debug purposes. */
  public String toString() {
    return (new Integer(this.hashCode())).toString();
  }

  
  public void update() {
    paint(getGraphics());
  }

}
