
//Title:        Indeco Steel Simulator
//Version:
//Copyright:    Copyright (c) 1999
//Author:       Joseph Guse & Ginger Booth
//Company:      Yale FES/CCE
//Description:  Steel industry economic simulator

/** PlotterTrace packages name, double[] array, color, etc. for one
 *  trace line on plot. Used for y-values on Plotter, x-automatic.
 */

/** Adopted from Indeco for Gecko-in-CourseWare, 990708, vfb. */

package stats;

import java.awt.*;

public class PlotterTrace {
  Color color;
  String name;
  int begin = 0;
  protected double[] values;
  public final static Color defaultColor = Color.black;
  public final static double[] defaultValues = new double[] { 0.0, 0.0 } ;
  public final static String defaultName = "uninit";

  int npts;

  public PlotterTrace() {
    this(defaultName, defaultColor, defaultValues );
  }
  public PlotterTrace(String name) {
    this(name, defaultColor, defaultValues );
  }
  public PlotterTrace(String name, Color color) {
    this(name, color, defaultValues);
  }
  public PlotterTrace(String name, Color color, double[] values) {
    this.name = name;
    this.values = values;
    this.color = color;
    npts = values.length;
  }

  public void clear() {
    values = defaultValues;
    npts = values.length;
  }
  /** There's really a choice here, either setting the trace to any-old array
   *  of doubles, OR giving it values one by one. This is the any-old array one.
   */
  public void setValues(double[] values) {
    this.values=values;
    if (values == null) npts = 0;
    else npts=values.length;
  }
  public void setColor(Color c) {color = c;}
  public void setName(String s) {name = s;}
	public String getName(){return name;}

  public int getNpts() {return npts;}

  /** Can optionally have a "begin" value stored. Like, all points begin there.
   *  That means, pt[i] corresponds to some external pt[begin+i].  */
  public int getBegin() { return begin; }
  public void setBegin(int i) { begin = i; }
  public double getVal(int i) {
    if (i>=npts)
      return Double.NaN;
    return values[i];
  }
  /** get latest value */
  public double getVal() {
    return values[npts-1];
  }
  public double[] getValues() {return values;}

  /** Doubles the array length. Does not change npts. */
  public void extend() {
    int newlen = values.length * 2;
    if (newlen < 10) newlen = 10;
    double[] newvalues = new double[newlen];
    for (int i=0; i<values.length; i++) {
      newvalues[i] = values[i];
    }
    values = newvalues;
  }
  /** There's really a choice here, EITHER setting the trace to any-old array
   *  of doubles, OR giving it values one by one. This is the one by one.
   */
  public void add(double value) {
    npts++;
    if (npts > values.length)
      extend();
    values[npts-1] = value;
  }
  public void set(int i, double value) {
    if (i > values.length - 1) extend();
    npts = npts > i+1 ? npts : i+1;
    values[i] = value;
  }
  /** stepping a data-only trace does nothing. */
  public void step() {
System.err.println("PlotterTrace steps, this=" + this);
  }
}